"""The code for the Agenta SDK"""
import argparse
import functools
import inspect
import os
import sys
import traceback
from typing import Any, Callable, Optional

from dotenv import load_dotenv
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

app = FastAPI()

origins = [
    "http://localhost:3000",
    "http://localhost:3001",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


class TextParam(str):
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update({"x-parameter": "text"})


class FloatParam(float):
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update({"x-parameter": "float"})


def post(func: Callable[..., Any]):
    load_dotenv()  # TODO: remove later when we have a better way to inject env variables
    sig = inspect.signature(func)
    func_params = sig.parameters

    # find the optional parameters for the app
    app_params = {
        name: param
        for name, param in func_params.items()
        if param.annotation in {TextParam, FloatParam}
    }
    # find the default values for the optional parameters
    for name, param in app_params.items():
        default_value = param.default if param.default is not param.empty else None
        app_params[name] = default_value

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs = {**app_params, **kwargs}
        try:
            return func(*args, **kwargs)
        except Exception as e:
            traceback_str = "".join(
                traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)
            )
            return JSONResponse(
                status_code=500, content={"error": str(e), "traceback": traceback_str}
            )

    new_params = []
    for name, param in sig.parameters.items():
        if name in app_params:
            new_params.append(
                inspect.Parameter(
                    name,
                    inspect.Parameter.KEYWORD_ONLY,
                    default=app_params[name],
                    annotation=Optional[param.annotation],
                )
            )
        else:
            new_params.append(param)

    wrapper.__signature__ = sig.replace(parameters=new_params)

    route = "/generate"
    app.post(route)(wrapper)

    # check if the module is being run as the main script
    if (
        os.path.splitext(os.path.basename(sys.argv[0]))[0]
        == os.path.splitext(os.path.basename(inspect.getfile(func)))[0]
    ):
        parser = argparse.ArgumentParser()
        # add arguments to the command-line parser
        for name, param in sig.parameters.items():
            if name in app_params:
                # For optional parameters, we add them as options
                parser.add_argument(
                    f"--{name}", type=type(param.default), default=param.default
                )
            else:
                # For required parameters, we add them as arguments
                parser.add_argument(name, type=param.annotation)

        args = parser.parse_args()
        print(func(**vars(args)))

    return wrapper



from chainlo.utils import check_module_version

if not check_module_version("langflow", "0.1.4"):
    raise ValueError(
        "Expected Langflow version >= 0.1.4. Run `pip install langflow --upgrade`"
    )

from typing import Dict, Optional, Union

import httpx
from chainlo.telemetry import trace_event


async def load_flow(schema: Union[Dict, str], tweaks: Optional[Dict] = None):
    from langflow import load_flow_from_json

    trace_event("load_langflow")

    if isinstance(schema, str):
        async with httpx.AsyncClient() as client:
            response = await client.get(schema)
            if response.status_code != 200:
                raise ValueError(f"Error: {response.text}")
            schema = response.json()

    flow = load_flow_from_json(flow=schema, tweaks=tweaks)

    return flow



from .utils.preinit import PreInitObject  # always the first import!
from . import agenta_decorator, context, types, utils  # noqa: F401
from .agenta_decorator import app, entrypoint
from .context import get_contexts, save_context
from .types import (
    Context,
    DictInput,
    FloatParam,
    InFile,
    IntParam,
    MultipleChoiceParam,
    TextParam,
    MessagesInput,
    FileInputURL,
    BinaryParam,
)
from .agenta_init import Config, init
from .utils.helper.openai_cost import calculate_token_usage

config = PreInitObject("agenta.config", Config)



from pydantic import BaseModel
from typing import List, Optional, Dict, Any


class AppVariant(BaseModel):
    app_id: str
    app_name: str
    variant_name: str
    variant_id: str
    parameters: Optional[Dict[str, Any]]
    previous_variant_name: Optional[str]
    base_name: Optional[str]
    config_name: Optional[str]


class Variant(BaseModel):
    variant_id: str


class Image(BaseModel):
    type: Optional[str]
    docker_id: str
    tags: str


class URI(BaseModel):
    uri: str


class VariantConfigPayload(BaseModel):
    base_id: str
    config_name: str
    parameters: Dict[str, Any]
    overwrite: bool



# This file was auto-generated by Fern from our API Definition.

from .types import (
    AddVariantFromBaseAndConfigResponse,
    AggregatedResult,
    App,
    AppVariantOutput,
    AppVariantOutputExtended,
    AppVariantRevision,
    BaseOutput,
    BodyImportTestset,
    ConfigDb,
    ContainerTemplatesResponse,
    CreateAppOutput,
    DeleteEvaluation,
    DockerEnvVars,
    EnvironmentOutput,
    Evaluation,
    EvaluationScenario,
    EvaluationScenarioInput,
    EvaluationScenarioOutput,
    EvaluationScenarioResult,
    EvaluationStatusEnum,
    EvaluationType,
    EvaluationWebhook,
    Evaluator,
    EvaluatorConfig,
    Feedback,
    GetConfigReponse,
    HttpValidationError,
    HumanEvaluation,
    HumanEvaluationScenario,
    HumanEvaluationScenarioInput,
    HumanEvaluationScenarioOutput,
    HumanEvaluationScenarioScore,
    HumanEvaluationScenarioUpdateScore,
    Image,
    InviteRequest,
    ListApiKeysOutput,
    LlmRunRateLimit,
    NewTestset,
    Organization,
    OrganizationOutput,
    Result,
    SimpleEvaluationOutput,
    Span,
    Template,
    TemplateImageInfo,
    TestSetOutputResponse,
    TestSetSimpleResponse,
    Trace,
    Uri,
    ValidationError,
    ValidationErrorLocItem,
    VariantAction,
    VariantActionEnum,
)
from .errors import UnprocessableEntityError

__all__ = [
    "AddVariantFromBaseAndConfigResponse",
    "AggregatedResult",
    "App",
    "AppVariantOutput",
    "AppVariantOutputExtended",
    "AppVariantRevision",
    "BaseOutput",
    "BodyImportTestset",
    "ConfigDb",
    "ContainerTemplatesResponse",
    "CreateAppOutput",
    "DeleteEvaluation",
    "DockerEnvVars",
    "EnvironmentOutput",
    "Evaluation",
    "EvaluationScenario",
    "EvaluationScenarioInput",
    "EvaluationScenarioOutput",
    "EvaluationScenarioResult",
    "EvaluationStatusEnum",
    "EvaluationType",
    "EvaluationWebhook",
    "Evaluator",
    "EvaluatorConfig",
    "Feedback",
    "GetConfigReponse",
    "HttpValidationError",
    "HumanEvaluation",
    "HumanEvaluationScenario",
    "HumanEvaluationScenarioInput",
    "HumanEvaluationScenarioOutput",
    "HumanEvaluationScenarioScore",
    "HumanEvaluationScenarioUpdateScore",
    "Image",
    "InviteRequest",
    "ListApiKeysOutput",
    "LlmRunRateLimit",
    "NewTestset",
    "Organization",
    "OrganizationOutput",
    "Result",
    "SimpleEvaluationOutput",
    "Span",
    "Template",
    "TemplateImageInfo",
    "TestSetOutputResponse",
    "TestSetSimpleResponse",
    "Trace",
    "UnprocessableEntityError",
    "Uri",
    "ValidationError",
    "ValidationErrorLocItem",
    "VariantAction",
    "VariantActionEnum",
]



from typing import List

from chainlo.context import context
from chainlo.input_widget import InputWidget
from pydantic.dataclasses import Field, dataclass


@dataclass
class ChatSettings:
    """Useful to create chat settings that the user can change."""

    inputs: List[InputWidget] = Field(default_factory=list, exclude=True)

    def __init__(
        self,
        inputs: List[InputWidget],
    ) -> None:
        self.inputs = inputs

    def settings(self):
        return dict(
            [(input_widget.id, input_widget.initial) for input_widget in self.inputs]
        )

    async def send(self):
        settings = self.settings()
        context.emitter.set_chat_settings(settings)

        inputs_content = [input_widget.to_dict() for input_widget in self.inputs]
        await context.emitter.emit("chat_settings", inputs_content)

        return settings



# https://raw.githubusercontent.com/langchain-ai/langchain/23eb480c3866db8693a3a2d63b787c898c54bb35/libs/community/langchain_community/callbacks/openai_info.py
MODEL_COST_PER_1K_TOKENS = {
    # GPT-4 input
    "gpt-4": 0.03,
    "gpt-4-0314": 0.03,
    "gpt-4-0613": 0.03,
    "gpt-4-32k": 0.06,
    "gpt-4-32k-0314": 0.06,
    "gpt-4-32k-0613": 0.06,
    "gpt-4-vision-preview": 0.01,
    "gpt-4-1106-preview": 0.01,
    # GPT-4 output
    "gpt-4-completion": 0.06,
    "gpt-4-0314-completion": 0.06,
    "gpt-4-0613-completion": 0.06,
    "gpt-4-32k-completion": 0.12,
    "gpt-4-32k-0314-completion": 0.12,
    "gpt-4-32k-0613-completion": 0.12,
    "gpt-4-vision-preview-completion": 0.03,
    "gpt-4-1106-preview-completion": 0.03,
    # GPT-3.5 input
    "gpt-3.5-turbo": 0.0015,
    "gpt-3.5-turbo-0301": 0.0015,
    "gpt-3.5-turbo-0613": 0.0015,
    "gpt-3.5-turbo-1106": 0.001,
    "gpt-3.5-turbo-instruct": 0.0015,
    "gpt-3.5-turbo-16k": 0.003,
    "gpt-3.5-turbo-16k-0613": 0.003,
    # GPT-3.5 output
    "gpt-3.5-turbo-completion": 0.002,
    "gpt-3.5-turbo-0301-completion": 0.002,
    "gpt-3.5-turbo-0613-completion": 0.002,
    "gpt-3.5-turbo-1106-completion": 0.002,
    "gpt-3.5-turbo-instruct-completion": 0.002,
    "gpt-3.5-turbo-16k-completion": 0.004,
    "gpt-3.5-turbo-16k-0613-completion": 0.004,
    # Azure GPT-35 input
    "gpt-35-turbo": 0.0015,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0301": 0.0015,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0613": 0.0015,
    "gpt-35-turbo-instruct": 0.0015,
    "gpt-35-turbo-16k": 0.003,
    "gpt-35-turbo-16k-0613": 0.003,
    # Azure GPT-35 output
    "gpt-35-turbo-completion": 0.002,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0301-completion": 0.002,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0613-completion": 0.002,
    "gpt-35-turbo-instruct-completion": 0.002,
    "gpt-35-turbo-16k-completion": 0.004,
    "gpt-35-turbo-16k-0613-completion": 0.004,
    # Others
    "text-ada-001": 0.0004,
    "ada": 0.0004,
    "text-babbage-001": 0.0005,
    "babbage": 0.0005,
    "text-curie-001": 0.002,
    "curie": 0.002,
    "text-davinci-003": 0.02,
    "text-davinci-002": 0.02,
    "code-davinci-002": 0.02,
    # Fine Tuned input
    "babbage-002-finetuned": 0.0016,
    "davinci-002-finetuned": 0.012,
    "gpt-3.5-turbo-0613-finetuned": 0.012,
    # Fine Tuned output
    "babbage-002-finetuned-completion": 0.0016,
    "davinci-002-finetuned-completion": 0.012,
    "gpt-3.5-turbo-0613-finetuned-completion": 0.016,
    # Azure Fine Tuned input
    "babbage-002-azure-finetuned": 0.0004,
    "davinci-002-azure-finetuned": 0.002,
    "gpt-35-turbo-0613-azure-finetuned": 0.0015,
    # Azure Fine Tuned output
    "babbage-002-azure-finetuned-completion": 0.0004,
    "davinci-002-azure-finetuned-completion": 0.002,
    "gpt-35-turbo-0613-azure-finetuned-completion": 0.002,
    # Legacy fine-tuned models
    "ada-finetuned-legacy": 0.0016,
    "babbage-finetuned-legacy": 0.0024,
    "curie-finetuned-legacy": 0.012,
    "davinci-finetuned-legacy": 0.12,
}


def standardize_model_name(
    model_name: str,
    is_completion: bool = False,
) -> str:
    """
    Standardize the model name to a format that can be used in the OpenAI API.

    Args:
        model_name: Model name to standardize.
        is_completion: Whether the model is used for completion or not.
            Defaults to False.

    Returns:
        Standardized model name.
    """

    model_name = model_name.lower()
    if ".ft-" in model_name:
        model_name = model_name.split(".ft-")[0] + "-azure-finetuned"
    if ":ft-" in model_name:
        model_name = model_name.split(":")[0] + "-finetuned-legacy"
    if "ft:" in model_name:
        model_name = model_name.split(":")[1] + "-finetuned"
    if is_completion and (
        model_name.startswith("gpt-4")
        or model_name.startswith("gpt-3.5")
        or model_name.startswith("gpt-35")
        or ("finetuned" in model_name and "legacy" not in model_name)
    ):
        return model_name + "-completion"
    else:
        return model_name


def get_openai_token_cost_for_model(
    model_name: str, num_tokens: int, is_completion: bool = False
) -> float:
    """
    Get the cost in USD for a given model and number of tokens.

    Args:
        model_name: Name of the model
        num_tokens: Number of tokens.
        is_completion: Whether the model is used for completion or not.
            Defaults to False.

    Returns:
        Cost in USD.
    """

    model_name = standardize_model_name(model_name, is_completion=is_completion)
    if model_name not in MODEL_COST_PER_1K_TOKENS:
        raise ValueError(
            f"Unknown model: {model_name}. Please provide a valid OpenAI model name."
            "Known models are: " + ", ".join(MODEL_COST_PER_1K_TOKENS.keys())
        )
    return MODEL_COST_PER_1K_TOKENS[model_name] * (num_tokens / 1000)


def calculate_token_usage(model_name: str, token_usage: dict) -> float:
    """Calculates the total cost of using a language model based on the model name and token
    usage.

    Args:
        model_name: The name of the model used to determine the cost per token.
        token_usage: Contains information about the usage of tokens for a particular model.

    Returns:
       Total cost of using a model.
    """

    completion_tokens = token_usage.get("completion_tokens", 0)
    prompt_tokens = token_usage.get("prompt_tokens", 0)
    model_name = standardize_model_name(model_name)
    if model_name in MODEL_COST_PER_1K_TOKENS:
        completion_cost = get_openai_token_cost_for_model(
            model_name, completion_tokens, is_completion=True
        )
        prompt_cost = get_openai_token_cost_for_model(model_name, prompt_tokens)
        total_cost = prompt_cost + completion_cost
        return total_cost
    return 0



import json
import time
from typing import Any, Dict, List, Optional, TypedDict, Union
from uuid import UUID

from chainlo.context import context_var
from chainlo.message import Message
from chainlo.step import Step
from langchain.callbacks.tracers.base import BaseTracer
from langchain.callbacks.tracers.schemas import Run
from langchain.schema import BaseMessage
from langchain.schema.output import ChatGenerationChunk, GenerationChunk
from langchain_core.outputs import ChatGenerationChunk, GenerationChunk
from literalai import ChatGeneration, CompletionGeneration, GenerationMessage
from literalai.helper import utc_now
from literalai.step import TrueStepType

DEFAULT_ANSWER_PREFIX_TOKENS = ["Final", "Answer", ":"]


class FinalStreamHelper:
    # The stream we can use to stream the final answer from a chain
    final_stream: Union[Message, None]
    # Should we stream the final answer?
    stream_final_answer: bool = False
    # Token sequence that prefixes the answer
    answer_prefix_tokens: List[str]
    # Ignore white spaces and new lines when comparing answer_prefix_tokens to last tokens? (to determine if answer has been reached)
    strip_tokens: bool

    answer_reached: bool

    def __init__(
        self,
        answer_prefix_tokens: Optional[List[str]] = None,
        stream_final_answer: bool = False,
        force_stream_final_answer: bool = False,
        strip_tokens: bool = True,
    ) -> None:
        # Langchain final answer streaming logic
        if answer_prefix_tokens is None:
            self.answer_prefix_tokens = DEFAULT_ANSWER_PREFIX_TOKENS
        else:
            self.answer_prefix_tokens = answer_prefix_tokens
        if strip_tokens:
            self.answer_prefix_tokens_stripped = [
                token.strip() for token in self.answer_prefix_tokens
            ]
        else:
            self.answer_prefix_tokens_stripped = self.answer_prefix_tokens

        self.last_tokens = [""] * len(self.answer_prefix_tokens)
        self.last_tokens_stripped = [""] * len(self.answer_prefix_tokens)
        self.strip_tokens = strip_tokens
        self.answer_reached = force_stream_final_answer

        # Our own final answer streaming logic
        self.stream_final_answer = stream_final_answer
        self.final_stream = None
        self.has_streamed_final_answer = False

    def _check_if_answer_reached(self) -> bool:
        if self.strip_tokens:
            return self._compare_last_tokens(self.last_tokens_stripped)
        else:
            return self._compare_last_tokens(self.last_tokens)

    def _compare_last_tokens(self, last_tokens: List[str]):
        if last_tokens == self.answer_prefix_tokens_stripped:
            # If tokens match perfectly we are done
            return True
        else:
            # Some LLMs will consider all the tokens of the final answer as one token
            # so we check if any last token contains all answer tokens
            return any(
                [
                    all(
                        answer_token in last_token
                        for answer_token in self.answer_prefix_tokens_stripped
                    )
                    for last_token in last_tokens
                ]
            )

    def _append_to_last_tokens(self, token: str) -> None:
        self.last_tokens.append(token)
        self.last_tokens_stripped.append(token.strip())
        if len(self.last_tokens) > len(self.answer_prefix_tokens):
            self.last_tokens.pop(0)
            self.last_tokens_stripped.pop(0)


class ChatGenerationStart(TypedDict):
    input_messages: List[BaseMessage]
    start: float
    token_count: int
    tt_first_token: Optional[float]


class CompletionGenerationStart(TypedDict):
    prompt: str
    start: float
    token_count: int
    tt_first_token: Optional[float]


class GenerationHelper:
    chat_generations: Dict[str, ChatGenerationStart]
    completion_generations: Dict[str, CompletionGenerationStart]
    generation_inputs: Dict[str, Dict]

    def __init__(self) -> None:
        self.chat_generations = {}
        self.completion_generations = {}
        self.generation_inputs = {}

    def ensure_values_serializable(self, data):
        """
        Recursively ensures that all values in the input (dict or list) are JSON serializable.
        """
        if isinstance(data, dict):
            return {
                key: self.ensure_values_serializable(value)
                for key, value in data.items()
            }
        elif isinstance(data, list):
            return [self.ensure_values_serializable(item) for item in data]
        elif isinstance(data, (str, int, float, bool, type(None))):
            return data
        elif isinstance(data, (tuple, set)):
            return list(data)  # Convert tuples and sets to lists
        else:
            return str(data)  # Fallback: convert other types to string

    def _convert_message_role(self, role: str):
        if "human" in role.lower():
            return "user"
        elif "system" in role.lower():
            return "system"
        elif "function" in role.lower():
            return "function"
        elif "tool" in role.lower():
            return "tool"
        else:
            return "assistant"

    def _convert_message_dict(
        self,
        message: Dict,
    ):
        class_name = message["id"][-1]
        kwargs = message.get("kwargs", {})
        function_call = kwargs.get("additional_kwargs", {}).get("function_call")

        msg = GenerationMessage(
            role=self._convert_message_role(class_name),
            content="",
        )
        if name := kwargs.get("name"):
            msg["name"] = name
        if function_call:
            msg["function_call"] = function_call
        else:
            msg["content"] = kwargs.get("content", "")

        return msg

    def _convert_message(
        self,
        message: Union[Dict, BaseMessage],
    ):
        if isinstance(message, dict):
            return self._convert_message_dict(
                message,
            )
        function_call = message.additional_kwargs.get("function_call")

        msg = GenerationMessage(
            role=self._convert_message_role(message.type),
            content="",
        )

        if literal_uuid := message.additional_kwargs.get("uuid"):
            msg["uuid"] = literal_uuid
            msg["templated"] = True

        if name := getattr(message, "name", None):
            msg["name"] = name

        if function_call:
            msg["function_call"] = function_call
        else:
            msg["content"] = message.content  # type: ignore

        return msg

    def _build_llm_settings(
        self,
        serialized: Dict,
        invocation_params: Optional[Dict] = None,
    ):
        # invocation_params = run.extra.get("invocation_params")
        if invocation_params is None:
            return None, None

        provider = invocation_params.pop("_type", "")  # type: str

        model_kwargs = invocation_params.pop("model_kwargs", {})

        if model_kwargs is None:
            model_kwargs = {}

        merged = {
            **invocation_params,
            **model_kwargs,
            **serialized.get("kwargs", {}),
        }

        # make sure there is no api key specification
        settings = {k: v for k, v in merged.items() if not k.endswith("_api_key")}

        model_keys = ["azure_deployment", "deployment_name", "model", "model_name"]
        model = next((settings[k] for k in model_keys if k in settings), None)
        tools = None
        if "functions" in settings:
            tools = [{"type": "function", "function": f} for f in settings["functions"]]
        if "tools" in settings:
            tools = settings["tools"]
        return provider, model, tools, settings


DEFAULT_TO_IGNORE = ["Runnable", "<lambda>"]
DEFAULT_TO_KEEP = ["retriever", "llm", "agent", "chain", "tool"]


class LangchainTracer(BaseTracer, GenerationHelper, FinalStreamHelper):
    steps: Dict[str, Step]
    parent_id_map: Dict[str, str]
    ignored_runs: set

    def __init__(
        self,
        # Token sequence that prefixes the answer
        answer_prefix_tokens: Optional[List[str]] = None,
        # Should we stream the final answer?
        stream_final_answer: bool = False,
        # Should force stream the first response?
        force_stream_final_answer: bool = False,
        # Runs to ignore to enhance readability
        to_ignore: Optional[List[str]] = None,
        # Runs to keep within ignored runs
        to_keep: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        BaseTracer.__init__(self, **kwargs)
        GenerationHelper.__init__(self)
        FinalStreamHelper.__init__(
            self,
            answer_prefix_tokens=answer_prefix_tokens,
            stream_final_answer=stream_final_answer,
            force_stream_final_answer=force_stream_final_answer,
        )
        self.context = context_var.get()
        self.steps = {}
        self.parent_id_map = {}
        self.ignored_runs = set()

        if self.context.current_step:
            self.root_parent_id = self.context.current_step.id
        elif self.context.session.root_message:
            self.root_parent_id = self.context.session.root_message.id
        else:
            self.root_parent_id = None

        if to_ignore is None:
            self.to_ignore = DEFAULT_TO_IGNORE
        else:
            self.to_ignore = to_ignore

        if to_keep is None:
            self.to_keep = DEFAULT_TO_KEEP
        else:
            self.to_keep = to_keep

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: "UUID",
        parent_run_id: Optional["UUID"] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        lc_messages = messages[0]
        self.chat_generations[str(run_id)] = {
            "input_messages": lc_messages,
            "start": time.time(),
            "token_count": 0,
            "tt_first_token": None,
        }

        return super().on_chat_model_start(
            serialized,
            messages,
            run_id=run_id,
            parent_run_id=parent_run_id,
            tags=tags,
            metadata=metadata,
            **kwargs,
        )

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: "UUID",
        tags: Optional[List[str]] = None,
        parent_run_id: Optional["UUID"] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> Run:
        self.completion_generations[str(run_id)] = {
            "prompt": prompts[0],
            "start": time.time(),
            "token_count": 0,
            "tt_first_token": None,
        }
        return super().on_llm_start(
            serialized,
            prompts,
            run_id=run_id,
            parent_run_id=parent_run_id,
            tags=tags,
            metadata=metadata,
            name=name,
            **kwargs,
        )

    def on_llm_new_token(
        self,
        token: str,
        *,
        chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
        run_id: "UUID",
        parent_run_id: Optional["UUID"] = None,
        **kwargs: Any,
    ) -> Run:
        if isinstance(chunk, ChatGenerationChunk):
            start = self.chat_generations[str(run_id)]
        else:
            start = self.completion_generations[str(run_id)]  # type: ignore
        start["token_count"] += 1
        if start["tt_first_token"] is None:
            start["tt_first_token"] = (time.time() - start["start"]) * 1000

        if self.stream_final_answer:
            self._append_to_last_tokens(token)

            if self.answer_reached:
                if not self.final_stream:
                    self.final_stream = Message(content="")
                    self._run_sync(self.final_stream.send())
                self._run_sync(self.final_stream.stream_token(token))
                self.has_streamed_final_answer = True
            else:
                self.answer_reached = self._check_if_answer_reached()

        return super().on_llm_new_token(
            token,
            chunk=chunk,
            run_id=run_id,
            parent_run_id=parent_run_id,
        )

    def _run_sync(self, co):  # TODO: WHAT TO DO WITH THIS?
        context_var.set(self.context)
        self.context.loop.create_task(co)

    def _persist_run(self, run: Run) -> None:
        pass

    def _get_run_parent_id(self, run: Run):
        parent_id = str(run.parent_run_id) if run.parent_run_id else self.root_parent_id

        return parent_id

    def _get_non_ignored_parent_id(self, current_parent_id: Optional[str] = None):
        if not current_parent_id:
            return self.root_parent_id

        if current_parent_id not in self.parent_id_map:
            return None

        while current_parent_id in self.parent_id_map:
            # If the parent id is in the ignored runs, we need to get the parent id of the ignored run
            if current_parent_id in self.ignored_runs:
                current_parent_id = self.parent_id_map[current_parent_id]
            else:
                return current_parent_id

        return self.root_parent_id

    def _should_ignore_run(self, run: Run):
        parent_id = self._get_run_parent_id(run)

        if parent_id:
            # Add the parent id of the ignored run in the mapping
            # so we can re-attach a kept child to the right parent id
            self.parent_id_map[str(run.id)] = parent_id

        ignore_by_name = False
        ignore_by_parent = parent_id in self.ignored_runs

        for filter in self.to_ignore:
            if filter in run.name:
                ignore_by_name = True
                break

        ignore = ignore_by_name or ignore_by_parent

        # If the ignore cause is the parent being ignored, check if we should nonetheless keep the child
        if ignore_by_parent and not ignore_by_name and run.run_type in self.to_keep:
            return False, self._get_non_ignored_parent_id(parent_id)
        else:
            if ignore:
                # Tag the run as ignored
                self.ignored_runs.add(str(run.id))
            return ignore, parent_id

    def _is_annotable(self, run: Run):
        return run.run_type in ["retriever", "llm"]

    def _start_trace(self, run: Run) -> None:
        super()._start_trace(run)
        context_var.set(self.context)

        ignore, parent_id = self._should_ignore_run(run)

        if run.run_type in ["chain", "prompt"]:
            self.generation_inputs[str(run.id)] = self.ensure_values_serializable(
                run.inputs
            )

        if ignore:
            return

        step_type: "TrueStepType" = "undefined"
        if run.run_type == "agent":
            step_type = "run"
        elif run.run_type == "chain":
            pass
        elif run.run_type == "llm":
            step_type = "llm"
        elif run.run_type == "retriever":
            step_type = "retrieval"
        elif run.run_type == "tool":
            step_type = "tool"
        elif run.run_type == "embedding":
            step_type = "embedding"

        if not self.steps:
            step_type = "run"

        disable_feedback = not self._is_annotable(run)

        step = Step(
            id=str(run.id),
            name=run.name,
            type=step_type,
            parent_id=parent_id,
            disable_feedback=disable_feedback,
        )
        step.start = utc_now()
        step.input = run.inputs

        self.steps[str(run.id)] = step

        self._run_sync(step.send())

    def _on_run_update(self, run: Run) -> None:
        """Process a run upon update."""
        context_var.set(self.context)

        ignore, parent_id = self._should_ignore_run(run)

        if ignore:
            return

        current_step = self.steps.get(str(run.id), None)

        if run.run_type == "llm" and current_step:
            provider, model, tools, llm_settings = self._build_llm_settings(
                (run.serialized or {}), (run.extra or {}).get("invocation_params")
            )
            generations = (run.outputs or {}).get("generations", [])
            generation = generations[0][0]
            variables = self.generation_inputs.get(str(run.parent_run_id), {})
            if message := generation.get("message"):
                chat_start = self.chat_generations[str(run.id)]
                duration = time.time() - chat_start["start"]
                if duration and chat_start["token_count"]:
                    throughput = chat_start["token_count"] / duration
                else:
                    throughput = None
                message_completion = self._convert_message(message)
                current_step.generation = ChatGeneration(
                    provider=provider,
                    model=model,
                    tools=tools,
                    variables=variables,
                    settings=llm_settings,
                    duration=duration,
                    token_throughput_in_s=throughput,
                    tt_first_token=chat_start.get("tt_first_token"),
                    messages=[
                        self._convert_message(m) for m in chat_start["input_messages"]
                    ],
                    message_completion=message_completion,
                )

                # find first message with prompt_id
                for m in chat_start["input_messages"]:
                    if m.additional_kwargs.get("prompt_id"):
                        current_step.generation.prompt_id = m.additional_kwargs[
                            "prompt_id"
                        ]
                        if custom_variables := m.additional_kwargs.get("variables"):
                            current_step.generation.variables = custom_variables
                    break

                current_step.language = "json"
                current_step.output = json.dumps(message_completion)
                completion = message_completion.get("content", "")
            else:
                completion_start = self.completion_generations[str(run.id)]
                completion = generation.get("text", "")
                duration = time.time() - completion_start["start"]
                if duration and completion_start["token_count"]:
                    throughput = completion_start["token_count"] / duration
                else:
                    throughput = None
                current_step.generation = CompletionGeneration(
                    provider=provider,
                    model=model,
                    settings=llm_settings,
                    variables=variables,
                    duration=duration,
                    token_throughput_in_s=throughput,
                    tt_first_token=completion_start.get("tt_first_token"),
                    prompt=completion_start["prompt"],
                    completion=completion,
                )
                current_step.output = completion

            if current_step:
                current_step.end = utc_now()
                self._run_sync(current_step.update())

            if self.final_stream and self.has_streamed_final_answer:
                self._run_sync(self.final_stream.update())

            return

        outputs = run.outputs or {}
        output_keys = list(outputs.keys())
        output = outputs
        if output_keys:
            output = outputs.get(output_keys[0], outputs)

        if current_step:
            current_step.output = output
            current_step.end = utc_now()
            self._run_sync(current_step.update())

    def _on_error(self, error: BaseException, *, run_id: UUID, **kwargs: Any):
        context_var.set(self.context)

        if current_step := self.steps.get(str(run_id), None):
            current_step.is_error = True
            current_step.output = str(error)
            current_step.end = utc_now()
            self._run_sync(current_step.update())

    on_llm_error = _on_error
    on_chain_error = _on_error
    on_tool_error = _on_error
    on_retriever_error = _on_error


LangchainCallbackHandler = LangchainTracer
AsyncLangchainCallbackHandler = LangchainTracer



# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class EvaluationStatusEnum(str, enum.Enum):
    """
    An enumeration.
    """

    EVALUATION_INITIALIZED = "EVALUATION_INITIALIZED"
    EVALUATION_STARTED = "EVALUATION_STARTED"
    EVALUATION_FINISHED = "EVALUATION_FINISHED"
    EVALUATION_FAILED = "EVALUATION_FAILED"

    def visit(
        self,
        evaluation_initialized: typing.Callable[[], T_Result],
        evaluation_started: typing.Callable[[], T_Result],
        evaluation_finished: typing.Callable[[], T_Result],
        evaluation_failed: typing.Callable[[], T_Result],
    ) -> T_Result:
        if self is EvaluationStatusEnum.EVALUATION_INITIALIZED:
            return evaluation_initialized()
        if self is EvaluationStatusEnum.EVALUATION_STARTED:
            return evaluation_started()
        if self is EvaluationStatusEnum.EVALUATION_FINISHED:
            return evaluation_finished()
        if self is EvaluationStatusEnum.EVALUATION_FAILED:
            return evaluation_failed()



from typing import Union

from chainlo.context import get_context
from chainlo.step import Step
from chainlo.sync import run_sync
from chainlo.utils import check_module_version
from literalai import ChatGeneration, CompletionGeneration
from literalai.helper import timestamp_utc


def instrument_openai():
    if not check_module_version("openai", "1.0.0"):
        raise ValueError(
            "Expected OpenAI version >= 1.0.0. Run `pip install openai --upgrade`"
        )

    from literalai.instrumentation.openai import instrument_openai

    async def on_new_generation(
        generation: Union["ChatGeneration", "CompletionGeneration"], timing
    ):
        context = get_context()

        parent_id = None
        if context.current_step:
            parent_id = context.current_step.id
        elif context.session.root_message:
            parent_id = context.session.root_message.id

        step = Step(
            name=generation.model if generation.model else generation.provider,
            type="llm",
            parent_id=parent_id,
        )
        step.generation = generation
        # Convert start/end time from seconds to milliseconds
        step.start = (
            timestamp_utc(timing.get("start"))
            if timing.get("start", None) is not None
            else None
        )
        step.end = (
            timestamp_utc(timing.get("end"))
            if timing.get("end", None) is not None
            else None
        )

        if isinstance(generation, ChatGeneration):
            step.input = generation.messages
            step.output = generation.message_completion  # type: ignore
        else:
            step.input = generation.prompt
            step.output = generation.completion

        await step.send()

    def on_new_generation_sync(
        generation: Union["ChatGeneration", "CompletionGeneration"], timing
    ):
        run_sync(on_new_generation(generation, timing))

    instrument_openai(None, on_new_generation_sync)



import os
from typing import Any, Dict, List, Optional, Union

from chainlo.config import config
from chainlo.telemetry import trace_event
from chainlo.types import GenerationRequest
from fastapi import HTTPException
from literalai import BaseGeneration, ChatGeneration, GenerationMessage
from pydantic.dataclasses import dataclass

from chainlo import input_widget


@dataclass
class BaseProvider:
    id: str
    name: str
    env_vars: Dict[str, str]
    inputs: List[input_widget.InputWidget]
    is_chat: bool

    # Convert the message to string format
    def message_to_string(self, message: GenerationMessage):
        return message["content"]

    # Concatenate multiple messages with a joiner
    def concatenate_messages(self, messages: List[GenerationMessage], joiner="\n\n"):
        return joiner.join([self.message_to_string(m) for m in messages])

    # Format the template based on the prompt inputs
    def _format_template(self, template: str, inputs: Optional[Dict]):
        return template.format(**(inputs or {}))

    # Create a prompt based on the request
    def create_generation(self, request: GenerationRequest):
        if request.chatGeneration and request.chatGeneration.messages:
            messages = request.chatGeneration.messages
        else:
            messages = None

        if self.is_chat:
            if messages:
                return messages
            elif request.completionGeneration and request.completionGeneration.prompt:
                return [
                    GenerationMessage(
                        content=request.completionGeneration.prompt,
                        role="user",
                    ),
                ]
            else:
                raise HTTPException(
                    status_code=422, detail="Could not create generation"
                )
        else:
            if request.completionGeneration:
                return request.completionGeneration.prompt
            elif messages:
                return self.concatenate_messages(messages)
            else:
                raise HTTPException(status_code=422, detail="Could not create prompt")

    # Create a completion event
    async def create_completion(self, request: GenerationRequest):
        trace_event("completion")

    # Get the environment variable based on the request
    def get_var(self, request: GenerationRequest, var: str) -> Union[str, None]:
        user_env = config.project.user_env or []

        if var in user_env:
            return request.userEnv.get(var)
        else:
            return os.environ.get(var)

    # Check if the environment variable is available
    def _is_env_var_available(self, var: str) -> bool:
        user_env = config.project.user_env or []
        return var in os.environ or var in user_env

    # Check if the provider is configured
    def is_configured(self):
        for var in self.env_vars.values():
            if not self._is_env_var_available(var):
                return False
        return True

    # Validate the environment variables in the request
    def validate_env(self, request: GenerationRequest):
        return {k: self.get_var(request, v) for k, v in self.env_vars.items()}

    # Check if the required settings are present
    def require_settings(self, settings: Dict[str, Any]):
        for _input in self.inputs:
            if _input.id not in settings:
                raise HTTPException(
                    status_code=422,
                    detail=f"Field {_input.id} is a required setting but is not found.",
                )

    # Convert the provider to dictionary format
    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "inputs": [input_widget.to_dict() for input_widget in self.inputs],
            "is_chat": self.is_chat,
        }



import chainlo.agenta_cli.agenta as ag
from langchain.chains import LLMChain
from langchain.llms import OpenAI
from langchain.prompts import PromptTemplate

import chainlo as cl


default_prompt = "What is a good name for a company that makes {product}?"

ag.init()
ag.config.default(
    temperature=ag.FloatParam(0.9),
    prompt_template=ag.TextParam(default_prompt),
)


@ag.entrypoint
def generate(product: str) -> str:
    llm = OpenAI(temperature=ag.config.temperature)
    prompt = PromptTemplate(
        input_variables=["product"], template=ag.config.prompt_template
    )
    chain = LLMChain(llm=llm, prompt=prompt)
    output = chain.run(product=product)

    return output


@cl.step
def tool():
    return "Response from the tool!"


@cl.on_message  # this function will be called every time a user inputs a message in the UI
async def main(message: cl.Message):
    """
    This function is called every time a user inputs a message in the UI.
    It sends back an intermediate response from the tool, followed by the final answer.

    Args:
        message: The user's message.

    Returns:
        None.
    """

    # Call the tool
    tool()

    # Send the final answer.
    await cl.Message(content="This is the final answer").send()


# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from .app_variant_revision import AppVariantRevision

try:
    import pydantic.v1 as pydantic  # type: ignore
except ImportError:
    import pydantic  # type: ignore


class AppVariantOutputExtended(pydantic.BaseModel):
    app_id: str
    app_name: str
    variant_id: str
    variant_name: str
    parameters: typing.Optional[typing.Dict[str, typing.Any]]
    previous_variant_name: typing.Optional[str]
    organization_id: str
    user_id: str
    base_name: str
    base_id: str
    config_name: str
    uri: typing.Optional[str]
    revision: int
    revisions: typing.List[AppVariantRevision]

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().dict(**kwargs_with_defaults)

    class Config:
        frozen = True
        smart_union = True
        json_encoders = {dt.datetime: serialize_datetime}



from typing import Any, Optional
from dotenv import load_dotenv

# from .context import setup_db

load_dotenv()
# setup_db()


class PreInitObject:
    """Dummy object that raises an error when accessed a class before agenta.init() is called."""

    def __init__(self, name: str, destination: Optional[Any] = None) -> None:
        self._name = name

        if destination is not None:
            self.__doc__ = destination.__doc__

    def __getitem__(self, key: str) -> None:
        raise RuntimeError(
            f"You must call agenta.init() before accessing {self._name}[{key!r}]"
        )

    def __setitem__(self, key: str, value: Any) -> Any:
        raise RuntimeError(
            f"You must call agenta.init() before setting {self._name}[{key!r}]"
        )

    def __setattr__(self, key: str, value: Any) -> Any:
        if not key.startswith("_"):
            raise RuntimeError(
                f"You must call agenta.init() before {self._name}[{key!r}]"
            )
        else:
            return object.__setattr__(self, key, value)

    def __getattr__(self, key: str) -> Any:
        if not key.startswith("_"):
            raise RuntimeError(f"You must call agenta.init() before {self._name}.{key}")
        else:
            raise AttributeError



# This file was auto-generated by Fern from our API Definition.

import typing

from .app_variant_output import AppVariantOutput

AddVariantFromBaseAndConfigResponse = typing.Union[AppVariantOutput, typing.Any]



from typing import Dict, Literal, TypedDict

from dataclasses_json import DataClassJsonMixin
from pydantic.dataclasses import Field, dataclass

Provider = Literal[
    "credentials", "header", "github", "google", "azure-ad", "okta", "auth0", "descope"
]


class UserDict(TypedDict):
    id: str
    identifier: str
    metadata: Dict


# Used when logging-in a user
@dataclass
class User(DataClassJsonMixin):
    identifier: str
    metadata: Dict = Field(default_factory=dict)


@dataclass
class PersistedUserFields:
    id: str
    createdAt: str


@dataclass
class PersistedUser(User, PersistedUserFields):
    pass



# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class VariantActionEnum(str, enum.Enum):
    """
    An enumeration.
    """

    START = "START"
    STOP = "STOP"

    def visit(
        self, start: typing.Callable[[], T_Result], stop: typing.Callable[[], T_Result]
    ) -> T_Result:
        if self is VariantActionEnum.START:
            return start()
        if self is VariantActionEnum.STOP:
            return stop()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from opentelemetry.sdk.resources import Attributes, Resource
    from opentelemetry.trace import Tracer

import hashlib
import logging
from functools import wraps
from socket import gethostname

from chainlo.config import config
from chainlo.version import __version__


class ChainlitTelemetry:
    def __init__(self):
        self._tracer = None

    # Patch uptrace.py to hash the hostname to avoid leaking it.
    @staticmethod
    def _build_resource(
        resource: "Resource",
        resource_attributes: "Attributes",
        service_name: str,
        service_version: str,
        deployment_environment: str,
    ) -> "Resource":
        from opentelemetry.sdk.resources import Resource

        if resource:
            return resource

        # Hash the local hostname to avoid leaking it.
        else:
            host_name = gethostname()
            host_name = hashlib.sha256(host_name.encode("UTF-8")).hexdigest()

        attrs = {"host.name": host_name}

        if resource_attributes:
            attrs.update(resource_attributes)  # type: ignore
        if service_name:
            attrs["service.name"] = service_name
        if service_version:
            attrs["service.version"] = service_version
        if deployment_environment:
            attrs["deployment.environment"] = deployment_environment

        return Resource.create(attrs)  # type: ignore

    def configure_tracer(self):
        import uptrace
        from opentelemetry.exporter.otlp.proto.grpc.exporter import (
            logger as exporter_logger,
        )
        from opentelemetry.trace import get_tracer

        if self._tracer:
            return self._tracer

        uptrace.uptrace._build_resource = self._build_resource

        uptrace.configure_opentelemetry(
            dsn="https://YPa4AbDF853uCW6UWN2oYg@api.uptrace.dev/1778",
            service_name="chainlit",
            deployment_environment="production",
            logging_level=logging.CRITICAL,
        )

        exporter_logger.setLevel(logging.CRITICAL)

        tracer = get_tracer("chainlit", __version__)
        return tracer

    @property
    def tracer(self) -> "Tracer":
        if self._tracer is None:
            self._tracer = self.configure_tracer()

        return self._tracer


chainlit_telemetry = ChainlitTelemetry()


def trace_event(event_name):
    if config.project.enable_telemetry:
        with chainlit_telemetry.tracer.start_as_current_span(
            event_name, record_exception=False, set_status_on_exception=False
        ):
            return


def trace(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if config.project.enable_telemetry:
            event_name = func.__name__
            with chainlit_telemetry.tracer.start_as_current_span(
                event_name, record_exception=False, set_status_on_exception=False
            ):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper



import click
from chainlo.agenta_cli.agenta.cli import helper


@click.group()
def config():
    """Commands for variants configurations"""
    pass


def update_backend_host(backend_host: str):
    """Check the config file and update the backend URL

    Arguments:
        app_folder -- the app folder
        backend_host -- the backend host
    """

    click.echo(
        click.style("\nChecking and updating global backend host...", fg="bright_black")
    )
    helper.set_global_config("host", backend_host)


@config.command(
    name="set-host",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.option(
    "--backend_host", default=None, help="The URL of the backend host to use."
)
@click.pass_context
def set_config_url(ctx, backend_host: str):
    """Set the backend URL in the app configuration"""

    try:
        if not backend_host:
            if ctx.args:
                backend_host = ctx.args[0]
            else:
                click.echo(click.style("Backend host URL not specified", fg="red"))

        update_backend_host(backend_host)
        click.echo(click.style("Backend host updated successfully! 🎉\n"))
    except Exception as ex:
        click.echo(click.style(f"Error updating backend host: {ex}", fg="red"))



import sys
from typing import Any, Coroutine, TypeVar

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

import asyncio
import threading

from asyncer import asyncify
from chainlo.context import context_var
from syncer import sync

make_async = asyncify

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")
T = TypeVar("T")


def run_sync(co: Coroutine[Any, Any, T_Retval]) -> T_Retval:
    """Run the coroutine synchronously."""

    # Copy the current context
    current_context = context_var.get()

    # Define a wrapper coroutine that sets the context before running the original coroutine
    async def context_preserving_coroutine():
        # Set the copied context to the coroutine
        context_var.set(current_context)
        return await co

    # Execute from the main thread in the main event loop
    if threading.current_thread() == threading.main_thread():
        return sync(context_preserving_coroutine())
    else:  # Execute from a thread in the main event loop
        result = asyncio.run_coroutine_threadsafe(
            context_preserving_coroutine(), loop=current_context.loop
        )
        return result.result()



import functools
import importlib
import inspect
from typing import Callable

from chainlo.context import context
from chainlo.logger import logger
from chainlo.message import ErrorMessage
from packaging import version


def wrap_user_function(user_function: Callable, with_task=False) -> Callable:
    """
    Wraps a user-defined function to accept arguments as a dictionary.

    Args:
        user_function (Callable): The user-defined function to wrap.

    Returns:
        Callable: The wrapped function.
    """

    @functools.wraps(user_function)
    async def wrapper(*args):
        # Get the parameter names of the user-defined function
        user_function_params = list(inspect.signature(user_function).parameters.keys())

        # Create a dictionary of parameter names and their corresponding values from *args
        params_values = {
            param_name: arg for param_name, arg in zip(user_function_params, args)
        }

        if with_task:
            await context.emitter.task_start()

        try:
            # Call the user-defined function with the arguments
            if inspect.iscoroutinefunction(user_function):
                return await user_function(**params_values)
            else:
                return user_function(**params_values)
        except InterruptedError:
            pass
        except Exception as e:
            logger.exception(e)
            await ErrorMessage(
                content=str(e) or e.__class__.__name__, author="Error"
            ).send()
        finally:
            if with_task:
                await context.emitter.task_end()

    return wrapper


def make_module_getattr(registry):
    """Leverage PEP 562 to make imports lazy in an __init__.py

    The registry must be a dictionary with the items to import as keys and the
    modules they belong to as a value.
    """

    def __getattr__(name):
        module_path = registry[name]
        module = importlib.import_module(module_path, __package__)
        return getattr(module, name)

    return __getattr__


def check_module_version(name, required_version):
    """
    Check the version of a module.

    Args:
        name (str): A module name.
        version (str): Minimum version.

    Returns:
        (bool): Return True if the module is installed and the version
            match the minimum required version.
    """
    try:
        module = importlib.import_module(name)
    except ModuleNotFoundError:
        return False
    return version.parse(module.__version__) >= version.parse(required_version)



import json
import sqlite3
from typing import List
from .types import Context


def setup_db():
    conn = sqlite3.connect("context.db")
    c = conn.cursor()
    c.execute(
        """
    CREATE TABLE IF NOT EXISTS contexts
    (id INTEGER PRIMARY KEY AUTOINCREMENT, context TEXT)
    """
    )
    conn.commit()
    conn.close()


def get_contexts() -> List[Context]:
    contexts = []
    conn = sqlite3.connect("context.db")
    c = conn.cursor()
    for row in c.execute("SELECT * FROM contexts"):
        context_data = json.loads(row[1])
        contexts.append(Context.parse_obj(context_data))
    conn.close()
    return contexts


def save_context(result: Context):
    conn = sqlite3.connect("context.db")
    c = conn.cursor()
    c.execute(
        """
    INSERT INTO contexts (context) VALUES (?)
    """,
        (json.dumps(result.dict()),),
    )
    conn.commit()
    conn.close()



import os

from dotenv import load_dotenv

env_found = load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

import asyncio
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

from fastapi import Request, Response
from pydantic.dataclasses import dataclass
from starlette.datastructures import Headers

if TYPE_CHECKING:
    from chainlo.haystack.callbacks import HaystackAgentCallbackHandler
    from chainlo.langchain.callbacks import (
        LangchainCallbackHandler,
        AsyncLangchainCallbackHandler,
    )
    from chainlo.llama_index.callbacks import LlamaIndexCallbackHandler
    from chainlo.openai import instrument_openai

import chainlo.input_widget as input_widget
from chainlo.action import Action
from chainlo.cache import cache
from chainlo.chat_settings import ChatSettings
from chainlo.config import config
from chainlo.context import context
from chainlo.element import (
    Audio,
    Avatar,
    File,
    Image,
    Pdf,
    Plotly,
    Pyplot,
    Task,
    TaskList,
    TaskStatus,
    Text,
    Video,
)
from chainlo.logger import logger
from chainlo.message import (
    AskActionMessage,
    AskFileMessage,
    AskUserMessage,
    ErrorMessage,
    Message,
)
from chainlo.oauth_providers import get_configured_oauth_providers
from chainlo.step import Step, step
from chainlo.sync import make_async, run_sync
from chainlo.telemetry import trace
from chainlo.types import ChatProfile, ThreadDict
from chainlo.user import PersistedUser, User
from chainlo.user_session import user_session
from chainlo.utils import make_module_getattr, wrap_user_function
from chainlo.version import __version__
from literalai import ChatGeneration, CompletionGeneration, GenerationMessage

if env_found:
    logger.info("Loaded .env file")


@trace
def password_auth_callback(func: Callable[[str, str], Optional[User]]) -> Callable:
    """
    Framework agnostic decorator to authenticate the user.

    Args:
        func (Callable[[str, str], Optional[User]]): The authentication callback to execute. Takes the email and password as parameters.

    Example:
        @cl.password_auth_callback
        async def password_auth_callback(username: str, password: str) -> Optional[User]:

    Returns:
        Callable[[str, str], Optional[User]]: The decorated authentication callback.
    """

    config.code.password_auth_callback = wrap_user_function(func)
    return func


@trace
def header_auth_callback(func: Callable[[Headers], Optional[User]]) -> Callable:
    """
    Framework agnostic decorator to authenticate the user via a header

    Args:
        func (Callable[[Headers], Optional[User]]): The authentication callback to execute.

    Example:
        @cl.header_auth_callback
        async def header_auth_callback(headers: Headers) -> Optional[User]:

    Returns:
        Callable[[Headers], Optional[User]]: The decorated authentication callback.
    """

    config.code.header_auth_callback = wrap_user_function(func)
    return func


@trace
def oauth_callback(
    func: Callable[[str, str, Dict[str, str], User], Optional[User]]
) -> Callable:
    """
    Framework agnostic decorator to authenticate the user via oauth

    Args:
        func (Callable[[str, str, Dict[str, str], User], Optional[User]]): The authentication callback to execute.

    Example:
        @cl.oauth_callback
        async def oauth_callback(provider_id: str, token: str, raw_user_data: Dict[str, str], default_app_user: User) -> Optional[User]:

    Returns:
        Callable[[str, str, Dict[str, str], User], Optional[User]]: The decorated authentication callback.
    """

    if len(get_configured_oauth_providers()) == 0:
        raise ValueError(
            "You must set the environment variable for at least one oauth provider to use oauth authentication."
        )

    config.code.oauth_callback = wrap_user_function(func)
    return func


@trace
def on_logout(func: Callable[[Request, Response], Any]) -> Callable:
    """
    Function called when the user logs out.
    Takes the FastAPI request and response as parameters.
    """

    config.code.on_logout = wrap_user_function(func)
    return func


@trace
def on_message(func: Callable) -> Callable:
    """
    Framework agnostic decorator to react to messages coming from the UI.
    The decorated function is called every time a new message is received.

    Args:
        func (Callable[[Message], Any]): The function to be called when a new message is received. Takes a cl.Message.

    Returns:
        Callable[[str], Any]: The decorated on_message function.
    """

    config.code.on_message = wrap_user_function(func)
    return func


@trace
def on_chat_start(func: Callable) -> Callable:
    """
    Hook to react to the user websocket connection event.

    Args:
        func (Callable[], Any]): The connection hook to execute.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.code.on_chat_start = wrap_user_function(func, with_task=True)
    return func


@trace
def on_chat_resume(func: Callable[[ThreadDict], Any]) -> Callable:
    """
    Hook to react to resume websocket connection event.

    Args:
        func (Callable[], Any]): The connection hook to execute.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.code.on_chat_resume = wrap_user_function(func, with_task=True)
    return func


@trace
def set_chat_profiles(
    func: Callable[[Optional["User"]], List["ChatProfile"]]
) -> Callable:
    """
    Programmatic declaration of the available chat profiles (can depend on the User from the session if authentication is setup).

    Args:
        func (Callable[[Optional["User"]], List["ChatProfile"]]): The function declaring the chat profiles.

    Returns:
        Callable[[Optional["User"]], List["ChatProfile"]]: The decorated function.
    """

    config.code.set_chat_profiles = wrap_user_function(func)
    return func


@trace
def on_chat_end(func: Callable) -> Callable:
    """
    Hook to react to the user websocket disconnect event.

    Args:
        func (Callable[], Any]): The disconnect hook to execute.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.code.on_chat_end = wrap_user_function(func, with_task=True)
    return func


@trace
def author_rename(func: Callable[[str], str]) -> Callable[[str], str]:
    """
    Useful to rename the author of message to display more friendly author names in the UI.
    Args:
        func (Callable[[str], str]): The function to be called to rename an author. Takes the original author name as parameter.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """

    config.code.author_rename = wrap_user_function(func)
    return func


@trace
def on_stop(func: Callable) -> Callable:
    """
    Hook to react to the user stopping a thread.

    Args:
        func (Callable[[], Any]): The stop hook to execute.

    Returns:
        Callable[[], Any]: The decorated stop hook.
    """

    config.code.on_stop = wrap_user_function(func)
    return func


def action_callback(name: str) -> Callable:
    """
    Callback to call when an action is clicked in the UI.

    Args:
        func (Callable[[Action], Any]): The action callback to execute. First parameter is the action.
    """

    def decorator(func: Callable[[Action], Any]):
        config.code.action_callbacks[name] = wrap_user_function(func, with_task=True)
        return func

    return decorator


def on_settings_update(
    func: Callable[[Dict[str, Any]], Any]
) -> Callable[[Dict[str, Any]], Any]:
    """
    Hook to react to the user changing any settings.

    Args:
        func (Callable[], Any]): The hook to execute after settings were changed.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.code.on_settings_update = wrap_user_function(func, with_task=True)
    return func


def sleep(duration: int):
    """
    Sleep for a given duration.
    Args:
        duration (int): The duration in seconds.
    """
    return asyncio.sleep(duration)


@dataclass()
class CopilotFunction:
    name: str
    args: Dict[str, Any]

    def acall(self):
        return context.emitter.send_call_fn(self.name, self.args)


__getattr__ = make_module_getattr(
    {
        "LangchainCallbackHandler": "chainlo.langchain.callbacks",
        "AsyncLangchainCallbackHandler": "chainlo.langchain.callbacks",
        "LlamaIndexCallbackHandler": "chainlo.llama_index.callbacks",
        "HaystackAgentCallbackHandler": "chainlo.haystack.callbacks",
        "instrument_openai": "chainlo.openai",
    }
)

__all__ = [
    "user_session",
    "CopilotFunction",
    "Action",
    "User",
    "PersistedUser",
    "Audio",
    "Pdf",
    "Plotly",
    "Image",
    "Text",
    "Avatar",
    "Pyplot",
    "File",
    "Task",
    "TaskList",
    "TaskStatus",
    "Video",
    "ChatSettings",
    "input_widget",
    "Message",
    "ErrorMessage",
    "AskUserMessage",
    "AskActionMessage",
    "AskFileMessage",
    "Step",
    "step",
    "ChatGeneration",
    "CompletionGeneration",
    "GenerationMessage",
    "on_logout",
    "on_chat_start",
    "on_chat_end",
    "on_chat_resume",
    "on_stop",
    "action_callback",
    "author_rename",
    "on_settings_update",
    "password_auth_callback",
    "header_auth_callback",
    "sleep",
    "run_sync",
    "make_async",
    "cache",
    "context",
    "LangchainCallbackHandler",
    "AsyncLangchainCallbackHandler",
    "LlamaIndexCallbackHandler",
    "HaystackAgentCallbackHandler",
    "instrument_openai",
]


def __dir__():
    return __all__



import asyncio
import json
import time
import uuid
from typing import Any, Dict, Literal

from chainlo.action import Action
from chainlo.auth import get_current_user, require_login
from chainlo.config import config
from chainlo.context import init_ws_context
from chainlo.data import get_data_layer
from chainlo.logger import logger
from chainlo.message import ErrorMessage, Message
from chainlo.server import socket
from chainlo.session import WebsocketSession
from chainlo.telemetry import trace_event
from chainlo.types import UIMessagePayload
from chainlo.user_session import user_sessions


def restore_existing_session(sid, session_id, emit_fn, emit_call_fn):
    """Restore a session from the sessionId provided by the client."""
    if session := WebsocketSession.get_by_id(session_id):
        session.restore(new_socket_id=sid)
        session.emit = emit_fn
        session.emit_call = emit_call_fn
        trace_event("session_restored")
        return True
    return False


async def persist_user_session(thread_id: str, metadata: Dict):
    if data_layer := get_data_layer():
        await data_layer.update_thread(thread_id=thread_id, metadata=metadata)


async def resume_thread(session: WebsocketSession):
    data_layer = get_data_layer()
    if not data_layer or not session.user or not session.thread_id_to_resume:
        return
    thread = await data_layer.get_thread(thread_id=session.thread_id_to_resume)
    if not thread:
        return

    author = thread.get("user").get("identifier") if thread["user"] else None
    user_is_author = author == session.user.identifier

    if user_is_author:
        metadata = thread.get("metadata", {})
        user_sessions[session.id] = metadata.copy()
        if chat_profile := metadata.get("chat_profile"):
            session.chat_profile = chat_profile
        if chat_settings := metadata.get("chat_settings"):
            session.chat_settings = chat_settings

        trace_event("thread_resumed")

        return thread


def load_user_env(user_env):
    # Check user env
    if config.project.user_env:
        # Check if requested user environment variables are provided
        if user_env:
            user_env = json.loads(user_env)
            for key in config.project.user_env:
                if key not in user_env:
                    trace_event("missing_user_env")
                    raise ConnectionRefusedError(
                        "Missing user environment variable: " + key
                    )
        else:
            raise ConnectionRefusedError("Missing user environment variables")
    return user_env


def build_anon_user_identifier(environ):
    scope = environ.get("asgi.scope", {})
    client_ip, _ = scope.get("client")
    ip = environ.get("HTTP_X_FORWARDED_FOR", client_ip)

    try:
        headers = scope.get("headers", {})
        user_agent = next(
            (v.decode("utf-8") for k, v in headers if k.decode("utf-8") == "user-agent")
        )
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, user_agent + ip))

    except StopIteration:
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, ip))


@socket.on("connect")
async def connect(sid, environ, auth):
    if not config.code.on_chat_start and not config.code.on_message:
        logger.warning(
            "You need to configure at least an on_chat_start or an on_message callback"
        )
        return False
    user = None
    token = None
    login_required = require_login()
    try:
        # Check if the authentication is required
        if login_required:
            authorization_header = environ.get("HTTP_AUTHORIZATION")
            token = authorization_header.split(" ")[1] if authorization_header else None
            user = await get_current_user(token=token)
    except Exception as e:
        logger.info("Authentication failed")
        return False

    # Session scoped function to emit to the client
    def emit_fn(event, data):
        if session := WebsocketSession.get(sid):
            if session.should_stop:
                session.should_stop = False
                raise InterruptedError("Task stopped by user")
        return socket.emit(event, data, to=sid)

    # Session scoped function to emit to the client and wait for a response
    def emit_call_fn(event: Literal["ask", "call_fn"], data, timeout):
        if session := WebsocketSession.get(sid):
            if session.should_stop:
                session.should_stop = False
                raise InterruptedError("Task stopped by user")
        return socket.call(event, data, timeout=timeout, to=sid)

    session_id = environ.get("HTTP_X_CHAINLIT_SESSION_ID")
    if restore_existing_session(sid, session_id, emit_fn, emit_call_fn):
        return True

    user_env_string = environ.get("HTTP_USER_ENV")
    user_env = load_user_env(user_env_string)

    client_type = environ.get("HTTP_X_CHAINLIT_CLIENT_TYPE")

    ws_session = WebsocketSession(
        id=session_id,
        socket_id=sid,
        emit=emit_fn,
        emit_call=emit_call_fn,
        client_type=client_type,
        user_env=user_env,
        user=user,
        token=token,
        chat_profile=environ.get("HTTP_X_CHAINLIT_CHAT_PROFILE"),
        thread_id=environ.get("HTTP_X_CHAINLIT_THREAD_ID"),
    )

    trace_event("connection_successful")
    return True


@socket.on("connection_successful")
async def connection_successful(sid):
    context = init_ws_context(sid)

    if context.session.restored:
        return

    await context.emitter.task_end()
    await context.emitter.clear("clear_ask")
    await context.emitter.clear("clear_call_fn")

    if context.session.thread_id_to_resume and config.code.on_chat_resume:
        thread = await resume_thread(context.session)
        if thread:
            context.session.has_first_interaction = True
            await context.emitter.emit("first_interaction", "resume")
            await context.emitter.resume_thread(thread)
            await config.code.on_chat_resume(thread)
            return

    if config.code.on_chat_start:
        await config.code.on_chat_start()


@socket.on("clear_session")
async def clean_session(sid):
    await disconnect(sid, force_clear=True)


@socket.on("disconnect")
async def disconnect(sid, force_clear=False):
    session = WebsocketSession.get(sid)
    if session:
        init_ws_context(session)

    if config.code.on_chat_end and session:
        await config.code.on_chat_end()

    if session and session.thread_id and session.has_first_interaction:
        await persist_user_session(session.thread_id, session.to_persistable())

    def clear():
        if session := WebsocketSession.get(sid):
            # Clean up the user session
            if session.id in user_sessions:
                user_sessions.pop(session.id)
            # Clean up the session
            session.delete()

    async def clear_on_timeout(sid):
        await asyncio.sleep(config.project.session_timeout)
        clear()

    if force_clear:
        clear()
    else:
        asyncio.ensure_future(clear_on_timeout(sid))


@socket.on("stop")
async def stop(sid):
    if session := WebsocketSession.get(sid):
        trace_event("stop_task")

        init_ws_context(session)
        await Message(
            author="System", content="Task stopped by the user.", disable_feedback=True
        ).send()

        session.should_stop = True

        if config.code.on_stop:
            await config.code.on_stop()


async def process_message(session: WebsocketSession, payload: UIMessagePayload):
    """Process a message from the user."""
    try:
        context = init_ws_context(session)
        await context.emitter.task_start()
        message = await context.emitter.process_user_message(payload)

        if config.code.on_message:
            # Sleep 1ms to make sure any children step starts after the message step start
            time.sleep(0.001)
            await config.code.on_message(message)
    except InterruptedError:
        pass
    except Exception as e:
        logger.exception(e)
        await ErrorMessage(
            author="Error", content=str(e) or e.__class__.__name__
        ).send()
    finally:
        await context.emitter.task_end()


@socket.on("ui_message")
async def message(sid, payload: UIMessagePayload):
    """Handle a message sent by the User."""
    session = WebsocketSession.require(sid)
    session.should_stop = False

    await process_message(session, payload)


async def process_action(action: Action):
    callback = config.code.action_callbacks.get(action.name)
    if callback:
        res = await callback(action)
        return res
    else:
        logger.warning("No callback found for action %s", action.name)


@socket.on("action_call")
async def call_action(sid, action):
    """Handle an action call from the UI."""
    context = init_ws_context(sid)

    action = Action(**action)

    try:
        res = await process_action(action)
        await context.emitter.send_action_response(
            id=action.id, status=True, response=res if isinstance(res, str) else None
        )

    except InterruptedError:
        await context.emitter.send_action_response(
            id=action.id, status=False, response="Action interrupted by the user"
        )
    except Exception as e:
        logger.exception(e)
        await context.emitter.send_action_response(
            id=action.id, status=False, response="An error occured"
        )


@socket.on("chat_settings_change")
async def change_settings(sid, settings: Dict[str, Any]):
    """Handle change settings submit from the UI."""
    context = init_ws_context(sid)

    for key, value in settings.items():
        context.session.chat_settings[key] = value

    if config.code.on_settings_update:
        await config.code.on_settings_update(settings)



from uvicorn import run
import chainlo.agenta_cli.agenta as agenta
import _app  # This will register the routes with the FastAPI application
import os

try:
    import ingest
except ImportError:
    pass


if __name__ == "__main__":
    run("agenta:app", host="0.0.0.0", port=80)



import uuid
from typing import Optional

from chainlo.context import context
from chainlo.telemetry import trace_event
from dataclasses_json import DataClassJsonMixin
from pydantic.dataclasses import Field, dataclass


@dataclass
class Action(DataClassJsonMixin):
    # Name of the action, this should be used in the action_callback
    name: str
    # The value associated with the action. This is useful to differentiate between multiple actions with the same name.
    value: str
    # The label of the action. This is what the user will see. If not provided the name will be used.
    label: str = ""
    # The description of the action. This is what the user will see when they hover the action.
    description: str = ""
    # This should not be set manually, only used internally.
    forId: Optional[str] = None
    # The ID of the action
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    # Show the action in a drawer menu
    collapsed: bool = False

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")

    async def send(self, for_id: str):
        trace_event(f"send {self.__class__.__name__}")
        self.forId = for_id
        await context.emitter.emit("action", self.to_dict())

    async def remove(self):
        trace_event(f"remove {self.__class__.__name__}")
        await context.emitter.emit("remove_action", self.to_dict())



import json
import mimetypes
import shutil
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Deque,
    Dict,
    List,
    Literal,
    Optional,
    Union,
)

import aiofiles
from chainlo.logger import logger

if TYPE_CHECKING:
    from chainlo.message import Message
    from chainlo.step import Step
    from chainlo.types import FileDict, FileReference
    from chainlo.user import PersistedUser, User

ClientType = Literal["app", "copilot", "teams", "slack"]


class JSONEncoderIgnoreNonSerializable(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(JSONEncoderIgnoreNonSerializable, self).default(obj)
        except TypeError:
            return None


def clean_metadata(metadata: Dict):
    return json.loads(
        json.dumps(metadata, cls=JSONEncoderIgnoreNonSerializable, ensure_ascii=False)
    )


class BaseSession:
    """Base object."""

    thread_id_to_resume: Optional[str] = None
    client_type: ClientType

    def __init__(
        self,
        # Id of the session
        id: str,
        client_type: ClientType,
        # Thread id
        thread_id: Optional[str],
        # Logged-in user informations
        user: Optional[Union["User", "PersistedUser"]],
        # Logged-in user token
        token: Optional[str],
        # User specific environment variables. Empty if no user environment variables are required.
        user_env: Optional[Dict[str, str]],
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # Chat profile selected before the session was created
        chat_profile: Optional[str] = None,
    ):
        if thread_id:
            self.thread_id_to_resume = thread_id
        self.thread_id = thread_id or str(uuid.uuid4())
        self.user = user
        self.client_type = client_type
        self.token = token
        self.root_message = root_message
        self.has_first_interaction = False
        self.user_env = user_env or {}
        self.chat_profile = chat_profile

        self.id = id

        self.chat_settings: Dict[str, Any] = {}

    async def persist_file(
        self,
        name: str,
        mime: str,
        path: Optional[str] = None,
        content: Optional[Union[bytes, str]] = None,
    ):
        return None

    def to_persistable(self) -> Dict:
        from chainlo.user_session import user_sessions

        user_session = user_sessions.get(self.id) or {}  # type: Dict
        user_session["chat_settings"] = self.chat_settings
        user_session["chat_profile"] = self.chat_profile
        metadata = clean_metadata(user_session)
        return metadata


class HTTPSession(BaseSession):
    """Internal HTTP session object. Used to consume Chainlit through API (no websocket)."""

    def __init__(
        self,
        # Id of the session
        id: str,
        client_type: ClientType,
        # Thread id
        thread_id: Optional[str] = None,
        # Logged-in user informations
        user: Optional[Union["User", "PersistedUser"]] = None,
        # Logged-in user token
        token: Optional[str] = None,
        user_env: Optional[Dict[str, str]] = None,
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # User specific environment variables. Empty if no user environment variables are required.
    ):
        super().__init__(
            id=id,
            thread_id=thread_id,
            user=user,
            token=token,
            client_type=client_type,
            user_env=user_env,
            root_message=root_message,
        )


class WebsocketSession(BaseSession):
    """Internal web socket session object.

    A socket id is an ephemeral id that can't be used as a session id
    (as it is for instance regenerated after each reconnection).

    The Session object store an internal mapping between socket id and
    a server generated session id, allowing to persists session
    between socket reconnection but also retrieving a session by
    socket id for convenience.
    """

    def __init__(
        self,
        # Id from the session cookie
        id: str,
        # Associated socket id
        socket_id: str,
        # Function to emit to the client
        emit: Callable[[str, Any], None],
        # Function to emit to the client and wait for a response
        emit_call: Callable[[Literal["ask", "call_fn"], Any, Optional[int]], Any],
        # User specific environment variables. Empty if no user environment variables are required.
        user_env: Dict[str, str],
        client_type: ClientType,
        # Thread id
        thread_id: Optional[str] = None,
        # Logged-in user informations
        user: Optional[Union["User", "PersistedUser"]] = None,
        # Logged-in user token
        token: Optional[str] = None,
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # Chat profile selected before the session was created
        chat_profile: Optional[str] = None,
    ):
        super().__init__(
            id=id,
            thread_id=thread_id,
            user=user,
            token=token,
            user_env=user_env,
            client_type=client_type,
            root_message=root_message,
            chat_profile=chat_profile,
        )

        self.socket_id = socket_id
        self.emit_call = emit_call
        self.emit = emit

        self.should_stop = False
        self.restored = False

        self.thread_queues = {}  # type: Dict[str, Deque[Callable]]
        self.files = {}  # type: Dict[str, "FileDict"]

        ws_sessions_id[self.id] = self
        ws_sessions_sid[socket_id] = self

    @property
    def files_dir(self):
        from chainlo.config import FILES_DIRECTORY

        return FILES_DIRECTORY / self.id

    async def persist_file(
        self,
        name: str,
        mime: str,
        path: Optional[str] = None,
        content: Optional[Union[bytes, str]] = None,
    ) -> "FileReference":
        if not path and not content:
            raise ValueError(
                "Either path or content must be provided to persist a file"
            )

        self.files_dir.mkdir(exist_ok=True)

        file_id = str(uuid.uuid4())

        file_path = self.files_dir / file_id

        file_extension = mimetypes.guess_extension(mime)
        if file_extension:
            file_path = file_path.with_suffix(file_extension)

        if path:
            # Copy the file from the given path
            async with aiofiles.open(path, "rb") as src, aiofiles.open(
                file_path, "wb"
            ) as dst:
                await dst.write(await src.read())
        elif content:
            # Write the provided content to the file
            async with aiofiles.open(file_path, "wb") as buffer:
                if isinstance(content, str):
                    content = content.encode("utf-8")
                await buffer.write(content)

        # Get the file size
        file_size = file_path.stat().st_size
        # Store the file content in memory
        self.files[file_id] = {
            "id": file_id,
            "path": file_path,
            "name": name,
            "type": mime,
            "size": file_size,
        }

        return {"id": file_id}

    def restore(self, new_socket_id: str):
        """Associate a new socket id to the session."""
        ws_sessions_sid.pop(self.socket_id, None)
        ws_sessions_sid[new_socket_id] = self
        self.socket_id = new_socket_id
        self.restored = True

    def delete(self):
        """Delete the session."""
        if self.files_dir.is_dir():
            shutil.rmtree(self.files_dir)
        ws_sessions_sid.pop(self.socket_id, None)
        ws_sessions_id.pop(self.id, None)

    async def flush_method_queue(self):
        for method_name, queue in self.thread_queues.items():
            while queue:
                method, self, args, kwargs = queue.popleft()
                try:
                    await method(self, *args, **kwargs)
                except Exception as e:
                    logger.error(f"Error while flushing {method_name}: {e}")

    @classmethod
    def get(cls, socket_id: str):
        """Get session by socket id."""
        return ws_sessions_sid.get(socket_id)

    @classmethod
    def get_by_id(cls, session_id: str):
        """Get session by session id."""
        return ws_sessions_id.get(session_id)

    @classmethod
    def require(cls, socket_id: str):
        """Throws an exception if the session is not found."""
        if session := cls.get(socket_id):
            return session
        raise ValueError("Session not found")


ws_sessions_sid: Dict[str, WebsocketSession] = {}
ws_sessions_id: Dict[str, WebsocketSession] = {}



import chainlo.agenta_cli.agenta as agenta


def set_global(setup=None, config=None):
    """Allows usage of agenta.config and agenta.setup in the user's code.

    Args:
        setup: _description_. Defaults to None.
        config: _description_. Defaults to None.
    """
    if setup is not None:
        agenta.setup = setup
    if config is not None:
        agenta.config = config



# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class EvaluationType(str, enum.Enum):
    """
    An enumeration.
    """

    HUMAN_A_B_TESTING = "human_a_b_testing"
    SINGLE_MODEL_TEST = "single_model_test"

    def visit(
        self,
        human_a_b_testing: typing.Callable[[], T_Result],
        single_model_test: typing.Callable[[], T_Result],
    ) -> T_Result:
        if self is EvaluationType.HUMAN_A_B_TESTING:
            return human_a_b_testing()
        if self is EvaluationType.SINGLE_MODEL_TEST:
            return single_model_test()



import os
import toml
import time
import click
from typing import Dict
from pathlib import Path
from chainlo.agenta_cli.agenta.client.backend import client
from chainlo.agenta_cli.agenta.client.api_models import Image
from requests.exceptions import RequestException
from chainlo.agenta_cli.agenta.client.backend.client import AgentaApi
from chainlo.agenta_cli.agenta.client.exceptions import APIRequestError


def add_variant_to_server(
    app_id: str,
    base_name: str,
    image: Image,
    backend_url: str,
    api_key: str,
    retries=10,
    backoff_factor=1,
) -> Dict:
    """
    Adds a variant to the server with a retry mechanism and a single-line loading state.

    Args:
        app_id (str): The ID of the app to add the variant to.
        base_name (str): The base name for the variant.
        image (Image): The image to use for the variant.
        retries (int): Number of times to retry the request.
        backoff_factor (float): Factor to determine the delay between retries (exponential backoff).

    Returns:
        dict: The JSON response from the server.

    Raises:
        APIRequestError: If the request to the server fails after retrying.
    """

    click.echo(
        click.style("Waiting for the variant to be ready", fg="yellow"), nl=False
    )

    client = AgentaApi(
        base_url=backend_url,
        api_key=api_key,
    )
    for attempt in range(retries):
        try:
            response = client.add_variant_from_image(
                app_id=app_id,
                variant_name=f"{base_name.lower()}.default",
                base_name=base_name,
                config_name="default",
                docker_id=image.docker_id,
                tags=image.tags,
            )
            click.echo(click.style("\nVariant added successfully.", fg="green"))
            return response
        except RequestException as e:
            if attempt < retries - 1:
                click.echo(click.style(".", fg="yellow"), nl=False)
                time.sleep(backoff_factor * (2**attempt))
            else:
                raise APIRequestError(
                    click.style(
                        f"\nRequest to app_variant endpoint failed with status code {response.status_code} and error message: {e}.",
                        fg="red",
                    )
                )
        except Exception as e:
            raise APIRequestError(
                click.style(f"\nAn unexpected error occurred: {e}", fg="red")
            )



from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, TypedDict, Union

if TYPE_CHECKING:
    from chainlo.element import ElementDict
    from chainlo.user import UserDict
    from chainlo.step import StepDict

from dataclasses_json import DataClassJsonMixin
from literalai import ChatGeneration, CompletionGeneration
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

InputWidgetType = Literal[
    "switch", "slider", "select", "textinput", "tags", "numberinput"
]


class ThreadDict(TypedDict):
    id: str
    createdAt: str
    name: Optional[str]
    user: Optional["UserDict"]
    tags: Optional[List[str]]
    metadata: Optional[Dict]
    steps: List["StepDict"]
    elements: Optional[List["ElementDict"]]


class Pagination(BaseModel):
    first: int
    cursor: Optional[str] = None


class ThreadFilter(BaseModel):
    feedback: Optional[Literal[-1, 0, 1]] = None
    userIdentifier: Optional[str] = None
    search: Optional[str] = None


@dataclass
class FileSpec(DataClassJsonMixin):
    accept: Union[List[str], Dict[str, List[str]]]
    max_files: int
    max_size_mb: int


@dataclass
class ActionSpec(DataClassJsonMixin):
    keys: List[str]


@dataclass
class AskSpec(DataClassJsonMixin):
    """Specification for asking the user."""

    timeout: int
    type: Literal["text", "file", "action"]


@dataclass
class AskFileSpec(FileSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user a file."""


@dataclass
class AskActionSpec(ActionSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user an action"""


class FileReference(TypedDict):
    id: str


class FileDict(TypedDict):
    id: str
    name: str
    path: str
    size: int
    type: str


class UIMessagePayload(TypedDict):
    message: "StepDict"
    fileReferences: Optional[List[FileReference]]


@dataclass
class AskFileResponse:
    id: str
    name: str
    path: str
    size: int
    type: str


class AskActionResponse(TypedDict):
    name: str
    value: str
    label: str
    description: str
    forId: str
    id: str
    collapsed: bool


class GenerationRequest(BaseModel):
    chatGeneration: Optional[ChatGeneration] = None
    completionGeneration: Optional[CompletionGeneration] = None
    userEnv: Dict[str, str]

    @property
    def generation(self):
        if self.chatGeneration:
            return self.chatGeneration
        return self.completionGeneration

    def is_chat(self):
        return self.chatGeneration is not None


class DeleteThreadRequest(BaseModel):
    threadId: str


class GetThreadsRequest(BaseModel):
    pagination: Pagination
    filter: ThreadFilter


class Theme(str, Enum):
    light = "light"
    dark = "dark"


@dataclass
class ChatProfile(DataClassJsonMixin):
    """Specification for a chat profile that can be chosen by the user at the thread start."""

    name: str
    markdown_description: str
    icon: Optional[str] = None


FeedbackStrategy = Literal["BINARY"]


class FeedbackDict(TypedDict):
    value: Literal[-1, 0, 1]
    strategy: FeedbackStrategy
    comment: Optional[str]


@dataclass
class Feedback:
    forId: str
    value: Literal[-1, 0, 1]
    strategy: FeedbackStrategy = "BINARY"
    id: Optional[str] = None
    comment: Optional[str] = None


class UpdateFeedbackRequest(BaseModel):
    feedback: Feedback



import json
import uuid
from enum import Enum
from io import BytesIO
from typing import Any, ClassVar, List, Literal, Optional, TypedDict, TypeVar, Union

import filetype
from chainlo.context import context
from chainlo.data import get_data_layer
from chainlo.logger import logger
from chainlo.telemetry import trace_event
from chainlo.types import FileDict
from pydantic.dataclasses import Field, dataclass
from syncer import asyncio

mime_types = {
    "text": "text/plain",
    "tasklist": "application/json",
    "plotly": "application/json",
}

ElementType = Literal[
    "image", "avatar", "text", "pdf", "tasklist", "audio", "video", "file", "plotly"
]
ElementDisplay = Literal["inline", "side", "page"]
ElementSize = Literal["small", "medium", "large"]


class ElementDict(TypedDict):
    id: str
    threadId: Optional[str]
    type: ElementType
    chainlitKey: Optional[str]
    url: Optional[str]
    objectKey: Optional[str]
    name: str
    display: ElementDisplay
    size: Optional[ElementSize]
    language: Optional[str]
    page: Optional[int]
    forId: Optional[str]
    mime: Optional[str]


@dataclass
class Element:
    # The type of the element. This will be used to determine how to display the element in the UI.
    type: ClassVar[ElementType]
    # Name of the element, this will be used to reference the element in the UI.
    name: str = ""
    # The ID of the element. This is set automatically when the element is sent to the UI.
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    # The key of the element hosted on Chainlit.
    chainlit_key: Optional[str] = None
    # The URL of the element if already hosted somehwere else.
    url: Optional[str] = None
    # The S3 object key.
    object_key: Optional[str] = None
    # The local path of the element.
    path: Optional[str] = None
    # The byte content of the element.
    content: Optional[Union[bytes, str]] = None
    # Controls how the image element should be displayed in the UI. Choices are “side” (default), “inline”, or “page”.
    display: ElementDisplay = Field(default="side")
    # Controls element size
    size: Optional[ElementSize] = None
    # The ID of the message this element is associated with.
    for_id: Optional[str] = None
    # The language, if relevant
    language: Optional[str] = None
    # Mime type, infered based on content if not provided
    mime: Optional[str] = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.persisted = False
        self.updatable = False
        self.thread_id = context.session.thread_id

        if not self.url and not self.path and not self.content:
            raise ValueError("Must provide url, path or content to instantiate element")

    def to_dict(self) -> ElementDict:
        _dict = ElementDict(
            {
                "id": self.id,
                "threadId": self.thread_id,
                "type": self.type,
                "url": self.url,
                "chainlitKey": self.chainlit_key,
                "name": self.name,
                "display": self.display,
                "objectKey": getattr(self, "object_key", None),
                "size": getattr(self, "size", None),
                "page": getattr(self, "page", None),
                "language": getattr(self, "language", None),
                "forId": getattr(self, "for_id", None),
                "mime": getattr(self, "mime", None),
            }
        )
        return _dict

    @classmethod
    def from_dict(self, _dict: FileDict):
        type = _dict.get("type", "")
        if "image" in type and "svg" not in type:
            return Image(
                id=_dict.get("id", str(uuid.uuid4())),
                name=_dict.get("name", ""),
                path=str(_dict.get("path")),
                chainlit_key=_dict.get("id"),
                display="inline",
                mime=type,
            )
        else:
            return File(
                id=_dict.get("id", str(uuid.uuid4())),
                name=_dict.get("name", ""),
                path=str(_dict.get("path")),
                chainlit_key=_dict.get("id"),
                display="inline",
                mime=type,
            )

    async def _create(self) -> bool:
        if self.persisted and not self.updatable:
            return True
        if data_layer := get_data_layer():
            try:
                asyncio.create_task(data_layer.create_element(self))
            except Exception as e:
                logger.error(f"Failed to create element: {str(e)}")
        if not self.url and (not self.chainlit_key or self.updatable):
            file_dict = await context.session.persist_file(
                name=self.name,
                path=self.path,
                content=self.content,
                mime=self.mime or "",
            )
            self.chainlit_key = file_dict["id"]

        self.persisted = True

        return True

    async def remove(self):
        trace_event(f"remove {self.__class__.__name__}")
        data_layer = get_data_layer()
        if data_layer and self.persisted:
            await data_layer.delete_element(self.id)
        await context.emitter.emit("remove_element", {"id": self.id})

    async def send(self, for_id: str):
        if self.persisted and not self.updatable:
            return

        self.for_id = for_id

        if not self.mime:
            # Only guess the mime type when the content is binary
            self.mime = (
                mime_types[self.type]
                if self.type in mime_types
                else filetype.guess_mime(self.path or self.content)
            )

        await self._create()

        if not self.url and not self.chainlit_key:
            raise ValueError("Must provide url or chainlit key to send element")

        trace_event(f"send {self.__class__.__name__}")
        await context.emitter.emit("element", self.to_dict())


ElementBased = TypeVar("ElementBased", bound=Element)


@dataclass
class Image(Element):
    type: ClassVar[ElementType] = "image"

    size: ElementSize = "medium"


@dataclass
class Avatar(Element):
    type: ClassVar[ElementType] = "avatar"

    async def send(self):
        await super().send(for_id="")


@dataclass
class Text(Element):
    """Useful to send a text (not a message) to the UI."""

    type: ClassVar[ElementType] = "text"
    language: Optional[str] = None


@dataclass
class Pdf(Element):
    """Useful to send a pdf to the UI."""

    mime: str = "application/pdf"
    page: Optional[int] = None
    type: ClassVar[ElementType] = "pdf"


@dataclass
class Pyplot(Element):
    """Useful to send a pyplot to the UI."""

    # We reuse the frontend image element to display the chart
    type: ClassVar[ElementType] = "image"

    size: ElementSize = "medium"
    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None

    def __post_init__(self) -> None:
        from matplotlib.figure import Figure

        if not isinstance(self.figure, Figure):
            raise TypeError("figure must be a matplotlib.figure.Figure")

        options = {
            "dpi": 200,
            "bbox_inches": "tight",
            "backend": "Agg",
            "format": "png",
        }
        image = BytesIO()
        self.figure.savefig(image, **options)
        self.content = image.getvalue()

        super().__post_init__()


class TaskStatus(Enum):
    READY = "ready"
    RUNNING = "running"
    FAILED = "failed"
    DONE = "done"


@dataclass
class Task:
    title: str
    status: TaskStatus = TaskStatus.READY
    forId: Optional[str] = None

    def __init__(
        self,
        title: str,
        status: TaskStatus = TaskStatus.READY,
        forId: Optional[str] = None,
    ):
        self.title = title
        self.status = status
        self.forId = forId


@dataclass
class TaskList(Element):
    type: ClassVar[ElementType] = "tasklist"
    tasks: List[Task] = Field(default_factory=list, exclude=True)
    status: str = "Ready"
    name: str = "tasklist"
    content: str = "dummy content to pass validation"

    def __post_init__(self) -> None:
        super().__post_init__()
        self.updatable = True

    async def add_task(self, task: Task):
        self.tasks.append(task)

    async def update(self):
        await self.send()

    async def send(self):
        await self.preprocess_content()
        await super().send(for_id="")

    async def preprocess_content(self):
        # serialize enum
        tasks = [
            {"title": task.title, "status": task.status.value, "forId": task.forId}
            for task in self.tasks
        ]

        # store stringified json in content so that it's correctly stored in the database
        self.content = json.dumps(
            {
                "status": self.status,
                "tasks": tasks,
            },
            indent=4,
            ensure_ascii=False,
        )


@dataclass
class Audio(Element):
    type: ClassVar[ElementType] = "audio"


@dataclass
class Video(Element):
    type: ClassVar[ElementType] = "video"

    size: ElementSize = "medium"


@dataclass
class File(Element):
    type: ClassVar[ElementType] = "file"


@dataclass
class Plotly(Element):
    """Useful to send a plotly to the UI."""

    type: ClassVar[ElementType] = "plotly"

    size: ElementSize = "medium"
    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None
    content: str = ""

    def __post_init__(self) -> None:
        from plotly import graph_objects as go
        from plotly import io as pio

        if not isinstance(self.figure, go.Figure):
            raise TypeError("figure must be a plotly.graph_objects.Figure")

        self.figure.layout.autosize = True
        self.figure.layout.width = None
        self.figure.layout.height = None
        self.content = pio.to_json(self.figure, validate=True)
        self.mime = "application/json"

        super().__post_init__()



import re
from typing import Any, Generic, List, Optional, TypeVar

from chainlo.context import context
from chainlo.step import Step
from chainlo.sync import run_sync
from haystack.agents import Agent, Tool
from haystack.agents.agent_step import AgentStep
from literalai.helper import utc_now

from chainlo import Message

T = TypeVar("T")


class Stack(Generic[T]):
    def __init__(self) -> None:
        self.items: List[T] = []

    def __len__(self):
        return len(self.items)

    def push(self, item: T) -> None:
        self.items.append(item)

    def pop(self) -> T:
        return self.items.pop()

    def peek(self) -> T:
        return self.items[-1]

    def clear(self) -> None:
        self.items.clear()


class HaystackAgentCallbackHandler:
    stack: Stack[Step]
    last_step: Optional[Step]

    def __init__(
        self,
        agent: Agent,
        stream_final_answer: bool = False,
        stream_final_answer_agent_name: str = "Agent",
    ):
        agent.callback_manager.on_agent_start += self.on_agent_start
        agent.callback_manager.on_agent_step += self.on_agent_step
        agent.callback_manager.on_agent_finish += self.on_agent_finish
        agent.callback_manager.on_new_token += self.on_new_token

        agent.tm.callback_manager.on_tool_start += self.on_tool_start
        agent.tm.callback_manager.on_tool_finish += self.on_tool_finish
        agent.tm.callback_manager.on_tool_error += self.on_tool_error

        self.final_answer_pattern = agent.final_answer_pattern
        self.stream_final_answer = stream_final_answer
        self.stream_final_answer_agent_name = stream_final_answer_agent_name

    def on_agent_start(self, **kwargs: Any) -> None:
        # Prepare agent step message for streaming
        self.agent_name = kwargs.get("name", "Agent")
        self.stack = Stack[Step]()

        if self.stream_final_answer:
            self.final_stream = Message(
                author=self.stream_final_answer_agent_name, content=""
            )
            self.last_tokens: List[str] = []
            self.answer_reached = False

        root_message = context.session.root_message
        parent_id = root_message.id if root_message else None
        run_step = Step(name=self.agent_name, type="run", parent_id=parent_id)
        run_step.start = utc_now()
        run_step.input = kwargs

        run_sync(run_step.send())

        self.stack.push(run_step)

    def on_agent_finish(self, agent_step: AgentStep, **kwargs: Any) -> None:
        if self.last_step:
            run_step = self.last_step
            run_step.end = utc_now()
            run_step.output = agent_step.prompt_node_response
            run_sync(run_step.update())

    # This method is called when a step has finished
    def on_agent_step(self, agent_step: AgentStep, **kwargs: Any) -> None:
        # Send previous agent step message
        self.last_step = self.stack.pop()

        # If token streaming is disabled
        if self.last_step.output == "":
            self.last_step.output = agent_step.prompt_node_response
        self.last_step.end = utc_now()
        run_sync(self.last_step.update())

        if not agent_step.is_last():
            # Prepare step for next agent step
            step = Step(name=self.agent_name, parent_id=self.last_step.id)
            self.stack.push(step)

    def on_new_token(self, token, **kwargs: Any) -> None:
        # Stream agent step tokens
        if self.stream_final_answer:
            if self.answer_reached:
                run_sync(self.final_stream.stream_token(token))
            else:
                self.last_tokens.append(token)

                last_tokens_concat = "".join(self.last_tokens)
                final_answer_match = re.search(
                    self.final_answer_pattern, last_tokens_concat
                )

                if final_answer_match:
                    self.answer_reached = True
                    run_sync(
                        self.final_stream.stream_token(final_answer_match.group(1))
                    )

        run_sync(self.stack.peek().stream_token(token))

    def on_tool_start(self, tool_input: str, tool: Tool, **kwargs: Any) -> None:
        # Tool started, create step
        parent_id = self.stack.items[0].id if self.stack.items[0] else None
        tool_step = Step(name=tool.name, type="tool", parent_id=parent_id)
        tool_step.input = tool_input
        tool_step.start = utc_now()
        self.stack.push(tool_step)

    def on_tool_finish(
        self,
        tool_result: str,
        tool_name: Optional[str] = None,
        tool_input: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        # Tool finished, send step with tool_result
        tool_step = self.stack.pop()
        tool_step.output = tool_result
        tool_step.end = utc_now()
        run_sync(tool_step.update())

    def on_tool_error(self, exception: Exception, tool: Tool, **kwargs: Any) -> None:
        # Tool error, send error message
        error_step = self.stack.pop()
        error_step.is_error = True
        error_step.output = str(exception)
        error_step.end = utc_now()
        run_sync(error_step.update())



from typing import Dict, Any, Optional
import os
import time
import click
from pathlib import Path
from typing import List, Optional, Dict, Any

import requests
from chainlo.agenta_cli.agenta.client.api_models import AppVariant, Image, VariantConfigPayload
from docker.models.images import Image as DockerImage
from requests.exceptions import RequestException

BACKEND_URL_SUFFIX = os.environ.get("BACKEND_URL_SUFFIX", "api")


class APIRequestError(Exception):
    """Exception to be raised when an API request fails."""


def get_base_by_app_id_and_name(
    app_id: str, base_name: str, host: str, api_key: str = None
) -> str:
    """
    Get the base ID for a given app ID and base name.

    Args:
        app_id (str): The ID of the app.
        base_name (str): The name of the base.
        host (str): The URL of the server.
        api_key (str, optional): The API key to use for authentication. Defaults to None.

    Returns:
        str: The ID of the base.

    Raises:
        APIRequestError: If the request to get the base fails or the base does not exist on the server.
    """
    response = requests.get(
        f"{host}/{BACKEND_URL_SUFFIX}/bases/?app_id={app_id}&base_name={base_name}",
        headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to get base failed with status code {response.status_code} and error message: {error_message}."
        )
    if len(response.json()) == 0:
        raise APIRequestError(
            f"Base with name {base_name} does not exist on the server."
        )
    else:
        return response.json()[0]["base_id"]


def get_app_by_name(app_name: str, host: str, api_key: str = None) -> str:
    """Get app by its name on the server.

    Args:
        app_name (str): Name of the app
        host (str): Hostname of the server
        api_key (str): The API key to use for the request.
    """

    response = requests.get(
        f"{host}/{BACKEND_URL_SUFFIX}/apps/?app_name={app_name}",
        headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to get app failed with status code {response.status_code} and error message: {error_message}."
        )
    if len(response.json()) == 0:
        raise APIRequestError(f"App with name {app_name} does not exist on the server.")
    else:
        return response.json()[0]["app_id"]  # only one app should exist for that name


def create_new_app(app_name: str, host: str, api_key: str = None) -> str:
    """Creates new app on the server.

    Args:
        app_name (str): Name of the app
        host (str): Hostname of the server
        api_key (str): The API key to use for the request.
    """

    response = requests.post(
        f"{host}/{BACKEND_URL_SUFFIX}/apps/",
        json={"app_name": app_name},
        headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to create new app failed with status code {response.status_code} and error message: {error_message}."
        )
    return response.json()["app_id"]


def add_variant_to_server(
    app_id: str,
    base_name: str,
    image: Image,
    host: str,
    api_key: str = None,
    retries=10,
    backoff_factor=1,
) -> Dict:
    """
    Adds a variant to the server with a retry mechanism and a single-line loading state.

    Args:
        app_id (str): The ID of the app to add the variant to.
        base_name (str): The base name for the variant.
        image (Image): The image to use for the variant.
        host (str): The host URL of the server.
        api_key (str): The API key to use for the request.
        retries (int): Number of times to retry the request.
        backoff_factor (float): Factor to determine the delay between retries (exponential backoff).

    Returns:
        dict: The JSON response from the server.

    Raises:
        APIRequestError: If the request to the server fails after retrying.
    """
    variant_name = f"{base_name.lower()}.default"
    payload = {
        "variant_name": variant_name,
        "base_name": base_name.lower(),
        "config_name": "default",
        "docker_id": image.docker_id,
        "tags": image.tags,
    }

    click.echo(
        click.style("Waiting for the variant to be ready", fg="yellow"), nl=False
    )

    for attempt in range(retries):
        try:
            response = requests.post(
                f"{host}/{BACKEND_URL_SUFFIX}/apps/{app_id}/variant/from-image/",
                json=payload,
                headers={"Authorization": api_key} if api_key is not None else None,
                timeout=600,
            )
            response.raise_for_status()
            click.echo(click.style("\nVariant added successfully.", fg="green"))
            return response.json()
        except RequestException as e:
            if attempt < retries - 1:
                click.echo(click.style(".", fg="yellow"), nl=False)
                time.sleep(backoff_factor * (2**attempt))
            else:
                raise APIRequestError(
                    click.style(
                        f"\nRequest to app_variant endpoint failed with status code {response.status_code} and error message: {e}.",
                        fg="red",
                    )
                )
        except Exception as e:
            raise APIRequestError(
                click.style(f"\nAn unexpected error occurred: {e}", fg="red")
            )


def start_variant(
    variant_id: str,
    host: str,
    env_vars: Optional[Dict[str, str]] = None,
    api_key: str = None,
) -> str:
    """
    Starts or stops a container with the given variant and exposes its endpoint.

    Args:
        variant_id (str): The ID of the variant.
        host (str): The host URL.
        env_vars (Optional[Dict[str, str]]): Optional environment variables to inject into the container.
        api_key (str): The API key to use for the request.

    Returns:
        str: The endpoint of the container.

    Raises:
        APIRequestError: If the API request fails.
    """
    payload = {}
    payload["action"] = {"action": "START"}
    if env_vars:
        payload["env_vars"] = env_vars
    try:
        response = requests.put(
            f"{host}/{BACKEND_URL_SUFFIX}/variants/{variant_id}/",
            json=payload,
            headers={"Authorization": api_key} if api_key is not None else None,
            timeout=600,
        )
        if response.status_code == 404:
            raise APIRequestError(
                f"404: Variant with ID {variant_id} does not exist on the server."
            )
        elif response.status_code != 200:
            error_message = response.text
            raise APIRequestError(
                f"Request to start variant endpoint failed with status code {response.status_code} and error message: {error_message}."
            )
        return response.json().get("uri", "")

    except RequestException as e:
        raise APIRequestError(f"An error occurred while making the request: {e}")


def list_variants(app_id: str, host: str, api_key: str = None) -> List[AppVariant]:
    """
    Returns a list of AppVariant objects for a given app_id and host.

    Args:
        app_id (str): The ID of the app to retrieve variants for.
        host (str): The URL of the host to make the request to.
        api_key (str): The API key to use for the request.

    Returns:
        List[AppVariant]: A list of AppVariant objects for the given app_id and host.
    """
    response = requests.get(
        f"{host}/{BACKEND_URL_SUFFIX}/apps/{app_id}/variants/",
        headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    # Check for successful request
    if response.status_code == 403:
        raise APIRequestError(
            f"No app by id {app_id} exists or you do not have access to it."
        )
    elif response.status_code == 404:
        raise APIRequestError(
            f"No app by id {app_id} exists or you do not have access to it."
        )
    elif response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to apps endpoint failed with status code {response.status_code} and error message: {error_message}."
        )

    app_variants = response.json()
    return [AppVariant(**variant) for variant in app_variants]


def remove_variant(variant_id: str, host: str, api_key: str = None):
    """
    Sends a DELETE request to the Agenta backend to remove a variant with the given ID.

    Args:
        variant_id (str): The ID of the variant to be removed.
        host (str): The URL of the Agenta backend.
        api_key (str): The API key to use for the request.

    Raises:
        APIRequestError: If the request to the remove_variant endpoint fails.

    Returns:
        None
    """
    response = requests.delete(
        f"{host}/{BACKEND_URL_SUFFIX}/variants/{variant_id}",
        headers={
            "Content-Type": "application/json",
            "Authorization": api_key if api_key is not None else None,
        },
        timeout=600,
    )

    # Check for successful request
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to remove_variant endpoint failed with status code {response.status_code} and error message: {error_message}"
        )


def update_variant_image(variant_id: str, image: Image, host: str, api_key: str = None):
    """
    Update the image of a variant with the given ID.

    Args:
        variant_id (str): The ID of the variant to update.
        image (Image): The new image to set for the variant.
        host (str): The URL of the host to send the request to.
        api_key (str): The API key to use for the request.

    Raises:
        APIRequestError: If the request to update the variant fails.

    Returns:
        None
    """
    response = requests.put(
        f"{host}/{BACKEND_URL_SUFFIX}/variants/{variant_id}/image/",
        json=image.dict(),
        headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to update app_variant failed with status code {response.status_code} and error message: {error_message}."
        )


def send_docker_tar(
    app_id: str, base_name: str, tar_path: Path, host: str, api_key: str = None
) -> Image:
    """
    Sends a Docker tar file to the specified host to build an image for the given app ID and variant name.

    Args:
        app_id (str): The ID of the app.
        base_name (str): The name of the codebase.
        tar_path (Path): The path to the Docker tar file.
        host (str): The URL of the host to send the request to.
        api_key (str): The API key to use for the request.

    Returns:
        Image: The built Docker image.

    Raises:
        Exception: If the response status code is 500, indicating that serving the variant failed.
    """
    with tar_path.open("rb") as tar_file:
        response = requests.post(
            f"{host}/{BACKEND_URL_SUFFIX}/containers/build_image/?app_id={app_id}&base_name={base_name}",
            files={
                "tar_file": tar_file,
            },
            headers={"Authorization": api_key} if api_key is not None else None,
            timeout=1200,
        )

    if response.status_code == 500:
        response_error = response.json()
        error_msg = "Serving the variant failed.\n"
        error_msg += f"Log: {response_error}\n"
        error_msg += "Here's how you may be able to solve the issue:\n"
        error_msg += "- First, make sure that the requirements.txt file has all the dependencies that you need.\n"
        error_msg += "- Second, check the Docker logs for the backend image to see the error when running the Docker container."
        raise Exception(error_msg)

    response.raise_for_status()
    image = Image.parse_obj(response.json())
    return image


def save_variant_config(
    base_id: str,
    config_name: str,
    parameters: Dict[str, Any],
    overwrite: bool,
    host: str,
    api_key: Optional[str] = None,
) -> None:
    """
    Saves a variant configuration to the Agenta backend.
    If the config already exists, it will be overwritten if the overwrite argument is set to True.
    If the config does does not exist, a new variant will be created.

    Args:
        base_id (str): The ID of the base configuration.
        config_name (str): The name of the variant configuration.
        parameters (Dict[str, Any]): The parameters of the variant configuration.
        overwrite (bool): Whether to overwrite an existing variant configuration with the same name.
        host (str): The URL of the Agenta backend.
        api_key (Optional[str], optional): The API key to use for authentication. Defaults to None.

    Raises:
        ValueError: If the 'host' argument is not specified.
        APIRequestError: If the request to the Agenta backend fails.

    Returns:
        None
    """
    if host is None:
        raise ValueError("The 'host' is not specified in save_variant_config")

    variant_config = VariantConfigPayload(
        base_id=base_id,
        config_name=config_name,
        parameters=parameters,
        overwrite=overwrite,
    )
    try:
        response = requests.post(
            f"{host}/{BACKEND_URL_SUFFIX}/configs/",
            json=variant_config.dict(),
            headers={"Authorization": api_key} if api_key is not None else None,
            timeout=600,
        )
        request = f"POST {host}/{BACKEND_URL_SUFFIX}/configs/ {variant_config.dict()}"
        # Check for successful request
        if response.status_code != 200:
            error_message = response.json().get("detail", "Unknown error")
            raise APIRequestError(
                f"Request {request} to save_variant_config endpoint failed with status code {response.status_code}. Error message: {error_message}"
            )
    except RequestException as e:
        raise APIRequestError(f"Request failed: {str(e)}")


def fetch_variant_config(
    base_id: str,
    host: str,
    config_name: Optional[str] = None,
    environment_name: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Fetch a variant configuration from the server.

    Args:
        base_id (str): ID of the base configuration.
        config_name (str): Configuration name.
        environment_name (str): Name of the environment.
        host (str): The server host URL.
        api_key (Optional[str], optional): The API key to use for authentication. Defaults to None.

    Raises:
        APIRequestError: If the API request fails.

    Returns:
        dict: The requested variant configuration.
    """

    if host is None:
        raise ValueError("The 'host' is not specified in fetch_variant_config")

    try:
        if environment_name:
            endpoint_params = f"?base_id={base_id}&environment_name={environment_name}"
        elif config_name:
            endpoint_params = f"?base_id={base_id}&config_name={config_name}"
        else:
            raise ValueError(
                "Either 'config_name' or 'environment_name' must be specified in fetch_variant_config"
            )
        response = requests.get(
            f"{host}/{BACKEND_URL_SUFFIX}/configs/{endpoint_params}",
            headers={"Authorization": api_key} if api_key is not None else None,
            timeout=600,
        )

        request = f"GET {host}/{BACKEND_URL_SUFFIX}/configs/ {base_id} {config_name} {environment_name}"

        # Check for successful request
        if response.status_code != 200:
            error_message = response.json().get("detail", "Unknown error")
            raise APIRequestError(
                f"Request {request} to fetch_variant_config endpoint failed with status code {response.status_code}. Error message: {error_message}"
            )

        return response.json()

    except RequestException as e:
        raise APIRequestError(f"Request failed: {str(e)}")


def validate_api_key(api_key: str, host: str) -> bool:
    """
    Validates an API key with the Agenta backend.

    Args:
        api_key (str): The API key to validate.
        host (str): The URL of the Agenta backend.

    Returns:
        bool: Whether the API key is valid or not.
    """
    try:
        headers = {"Authorization": api_key}

        prefix = api_key.split(".")[0]

        response = requests.get(
            f"{host}/{BACKEND_URL_SUFFIX}/keys/{prefix}/validate/",
            headers=headers,
            timeout=600,
        )
        if response.status_code != 200:
            error_message = response.json()
            raise APIRequestError(
                f"Request to validate api key failed with status code {response.status_code} and error message: {error_message}."
            )
        return True
    except RequestException as e:
        raise APIRequestError(f"An error occurred while making the request: {e}")


def retrieve_user_id(host: str, api_key: Optional[str] = None) -> str:
    """Retrieve user ID from the server.

    Args:
        host (str): The URL of the Agenta backend
        api_key (str): The API key to validate with.

    Returns:
        str: the user ID
    """

    try:
        response = requests.get(
            f"{host}/{BACKEND_URL_SUFFIX}/profile/",
            headers={"Authorization": api_key} if api_key is not None else None,
            timeout=600,
        )
        if response.status_code != 200:
            error_message = response.json().get("detail", "Unknown error")
            raise APIRequestError(
                f"Request to fetch_user_profile endpoint failed with status code {response.status_code}. Error message: {error_message}"
            )
        return response.json()["id"]
    except RequestException as e:
        raise APIRequestError(f"Request failed: {str(e)}")


from pydantic import BaseModel


# def run_evaluation(app_name: str, host: str, api_key: str = None) -> str:
def run_evaluation(app_name: str, host: str, api_key: str = None) -> str:
    """Creates new app on the server.
    Args:
        app_name (str): Name of the app
        host (str): Hostname of the server
        api_key (str): The API key to use for the request.
    """

    new_evaluation = {
        "app_id": "6583e552eb855930ec6b1bdd",
        "variant_ids": [
            "6583e552eb855930ec6b1be3",
            # "6570aed55d0eaff2293088e6"
        ],
        "evaluators_configs": ["65856b2b11d53fcce5894ab6"],
        "testset_id": "6583e552eb855930ec6b1be4",
    }

    response = requests.post(
        f"{host}/api/evaluations/",
        json=new_evaluation,
        # headers={"Authorization": api_key} if api_key is not None else None,
        timeout=600,
    )
    if response.status_code != 200:
        error_message = response.json()
        raise APIRequestError(
            f"Request to run evaluations failed with status code {response.status_code} and error message: {error_message}."
        )
    print(response.json())
    return response.json()



from chainlo.data import get_data_layer
from fastapi import HTTPException


async def is_thread_author(username: str, thread_id: str):
    data_layer = get_data_layer()
    if not data_layer:
        raise HTTPException(status_code=401, detail="Unauthorized")

    thread_author = await data_layer.get_thread_author(thread_id)

    if thread_author != username:
        raise HTTPException(status_code=401, detail="Unauthorized")
    else:
        return True



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from .feedback import Feedback

try:
    import pydantic.v1 as pydantic  # type: ignore
except ImportError:
    import pydantic  # type: ignore


class Trace(pydantic.BaseModel):
    app_id: typing.Optional[str]
    variant_id: typing.Optional[str]
    cost: typing.Optional[float]
    latency: float
    status: str
    token_consumption: typing.Optional[int]
    tags: typing.Optional[typing.List[str]]
    start_time: dt.datetime
    end_time: dt.datetime
    trace_id: str
    spans: typing.List[str]
    feedbacks: typing.Optional[typing.List[Feedback]]

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().dict(**kwargs_with_defaults)

    class Config:
        frozen = True
        smart_union = True
        json_encoders = {dt.datetime: serialize_datetime}



import json
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Extra, HttpUrl, Field


class InFile:
    def __init__(self, file_name: str, file_path: str):
        self.file_name = file_name
        self.file_path = file_path


class LLMTokenUsage(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int


class FuncResponse(BaseModel):
    message: str
    usage: Optional[LLMTokenUsage]
    cost: Optional[float]
    latency: float


class DictInput(dict):
    def __new__(cls, default_keys=None):
        instance = super().__new__(cls, default_keys)
        if default_keys is None:
            default_keys = []
        instance.data = [key for key in default_keys]
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update({"x-parameter": "dict"})


class TextParam(str):
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update({"x-parameter": "text"})


class BinaryParam(int):
    def __new__(cls, value: bool = False):
        instance = super().__new__(cls, int(value))
        instance.default = value
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            {
                "x-parameter": "bool",
                "type": "boolean",
            }
        )


class IntParam(int):
    def __new__(cls, default: int = 6, minval: float = 1, maxval: float = 10):
        instance = super().__new__(cls, default)
        instance.minval = minval
        instance.maxval = maxval
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            {
                "x-parameter": "int",
                "type": "integer",
                "minimum": 1,
                "maximum": 10,
            }
        )


class FloatParam(float):
    def __new__(cls, default: float = 0.5, minval: float = 0.0, maxval: float = 1.0):
        instance = super().__new__(cls, default)
        instance.minval = minval
        instance.maxval = maxval
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            {
                "x-parameter": "float",
                "type": "number",
                "minimum": 0.0,
                "maximum": 1.0,
            }
        )


class MultipleChoiceParam(str):
    def __new__(cls, default: str = None, choices: List[str] = None):
        if type(default) is list:
            raise ValueError(
                "The order of the parameters for MultipleChoiceParam is wrong! It's MultipleChoiceParam(default, choices) and not the opposite"
            )
        if default is None and choices:
            # if a default value is not provided,
            # uset the first value in the choices list
            default = choices[0]

        if default is None and not choices:
            # raise error if no default value or choices is provided
            raise ValueError("You must provide either a default value or choices")

        instance = super().__new__(cls, default)
        instance.choices = choices
        instance.default = default
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]):
        field_schema.update(
            {
                "x-parameter": "choice",
                "type": "string",
                "enum": [],
            }
        )


class Message(BaseModel):
    role: str
    content: str


class MessagesInput(list):
    """Messages Input for Chat-completion.

    Args:
        messages (List[Dict[str, str]]): The list of messages inputs.
        Required. Each message should be a dictionary with "role" and "content" keys.

    Raises:
        ValueError: If `messages` is not specified or empty.

    """

    def __new__(cls, messages: List[Dict[str, str]] = None):
        instance = super().__new__(cls, messages)
        instance.default = messages
        return instance

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]):
        field_schema.update({"x-parameter": "messages", "type": "array"})


class FileInputURL(HttpUrl):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        field_schema.update({"x-parameter": "file_url", "type": "string"})


class Context(BaseModel):
    class Config:
        extra = Extra.allow

    def to_json(self):
        return self.json()

    @classmethod
    def from_json(cls, json_str: str):
        data = json.loads(json_str)
        return cls(**data)



import os
import sys
import toml
import click
import questionary
from pathlib import Path
from typing import Any, List, MutableMapping
from chainlo.agenta_cli.agenta.client.api_models import AppVariant


from typing import Any, Optional
from pathlib import Path
import toml

from chainlo.agenta_cli.agenta.client.backend.client import AgentaApi

BACKEND_URL_SUFFIX = os.environ.get("BACKEND_URL_SUFFIX", "api")


def get_global_config(var_name: str) -> Optional[Any]:
    """
    Get the value of a global configuration variable.

    Args:
        var_name: the name of the variable to get

    Returns:
        the value of the variable, or None if it doesn't exist
    """
    agenta_dir = Path.home() / ".agenta"
    if not agenta_dir.exists():
        return None
    agenta_config_file = agenta_dir / "config.toml"
    if not agenta_config_file.exists():
        return None
    global_config = toml.load(agenta_config_file)
    if var_name not in global_config:
        return None
    return global_config[var_name]


def set_global_config(var_name: str, var_value: Any) -> None:
    """
    Set the value of a global configuration variable.

    Args:
        var_name: the name of the variable to set
        var_value: the value to set the variable to
    """
    agenta_dir = Path.home() / ".agenta"
    if not agenta_dir.exists():
        agenta_dir.mkdir(exist_ok=True)
    agenta_config_file = agenta_dir / "config.toml"
    if not agenta_config_file.exists():
        config = {}
        with agenta_config_file.open("w") as config_file:
            toml.dump(config, config_file)
    global_config = toml.load(agenta_config_file)
    global_config[var_name] = var_value
    with open(agenta_config_file, "w") as config_file:
        toml.dump(global_config, config_file)


def get_api_key(backend_host: str) -> str:
    """
    Retrieve or request the API key for accessing the Agenta platform.

    This function first looks for an existing API key in the global config file.
    If found, it prompts the user to confirm whether they'd like to use that key.
    If not found, it asks the user to input a new key.

    Args:
        backend_host (str): The URL of the backend host.

    Returns:
        str: The API key to be used for accessing the Agenta platform.

    Raises:
        SystemExit: If the user cancels the input by pressing Ctrl+C.
    """

    api_key = get_global_config("api_key")
    if api_key:
        # API key exists in the config file, ask for confirmation
        confirm_api_key = questionary.confirm(
            f"API Key found: {api_key}\nDo you want to use this API Key?"
        ).ask()

        if confirm_api_key:
            return api_key
        elif confirm_api_key is None:  # User pressed Ctrl+C
            sys.exit(0)

    api_key = questionary.text(
        f"(You can get your API Key here: {backend_host}/settings?tab=apiKeys) "
        "Please provide your API key:"
    ).ask()

    if api_key:
        set_global_config("api_key", api_key)

        return api_key
    elif api_key is None:  # User pressed Ctrl+C
        sys.exit(0)


def init_telemetry_config() -> None:
    if (
        get_global_config("telemetry_tracking_enabled") is None
        or get_global_config("telemetry_api_key") is None
    ):
        set_global_config("telemetry_tracking_enabled", True)
        set_global_config(
            "telemetry_api_key", "phc_hmVSxIjTW1REBHXgj2aw4HW9X6CXb6FzerBgP9XenC7"
        )


def update_variants_from_backend(
    app_id: str,
    config: MutableMapping[str, Any],
    host: str,
    api_key: str = None,
) -> MutableMapping[str, Any]:
    """Reads the list of variants from the backend and updates the config accordingly

    Arguments:
        app_id -- the app id
        config -- the config loaded using toml.load
        api_key -- the api key to use for authentication

    Returns:
        a new config object later to be saved using toml.dump(config, config_file.open('w'))
    """
    client = AgentaApi(
        base_url=f"{host}/{BACKEND_URL_SUFFIX}",
        api_key=api_key,
    )

    try:
        variants: List[AppVariant] = client.list_app_variants(app_id=app_id)
    except Exception as ex:
        raise ex

    config["variants"] = [variant.variant_name for variant in variants]
    config["variant_ids"] = [variant.variant_id for variant in variants]
    return config


def update_config_from_backend(config_file: Path, host: str):
    """Updates the config file with new information from the backend

    Arguments:
        config_file -- the path to the config file
    """
    assert config_file.exists(), "Config file does not exist!"
    config = toml.load(config_file)
    app_id = config["app_id"]
    api_key = config.get("api_key", "")
    if "variants" not in config:
        config["variants"] = []
    if "variant_ids" not in config:
        config["variant_ids"] = []
    config = update_variants_from_backend(app_id, config, host, api_key)
    toml.dump(config, config_file.open("w"))


def display_app_variant(variant: AppVariant):
    """Prints a variant nicely in the terminal"""
    click.echo(
        click.style("App Name: ", bold=True, fg="green")
        + click.style(variant.app_name, fg="green")
    )
    click.echo(
        click.style("Variant Name: ", bold=True, fg="blue")
        + click.style(variant.variant_name, fg="blue")
    )
    click.echo(click.style("Parameters: ", bold=True, fg="cyan"))
    if variant.parameters:
        for param, value in variant.parameters.items():
            click.echo(
                click.style(f"  {param}: ", fg="cyan")
                + click.style(str(value), fg="cyan")
            )
    else:
        click.echo(click.style("  Defaults from code", fg="cyan"))
    if variant.previous_variant_name:
        click.echo(
            click.style("Template Variant Name: ", bold=True, fg="magenta")
            + click.style(variant.previous_variant_name, fg="magenta")
        )
    else:
        click.echo(
            click.style("Template Variant Name: ", bold=True, fg="magenta")
            + click.style("None", fg="magenta")
        )
    click.echo(
        click.style("-" * 50, bold=True, fg="white")
    )  # a line for separating each variant



from abc import abstractmethod
from collections import defaultdict
from typing import Any, Dict, List, Optional

from chainlo.types import InputWidgetType
from pydantic.dataclasses import Field, dataclass


@dataclass
class InputWidget:
    id: str
    label: str
    initial: Any = None
    tooltip: Optional[str] = None
    description: Optional[str] = None

    def __post_init__(
        self,
    ) -> None:
        if not self.id or not self.label:
            raise ValueError("Must provide key and label to load InputWidget")

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass


@dataclass
class Switch(InputWidget):
    """Useful to create a switch input."""

    type: InputWidgetType = "switch"
    initial: bool = False

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Slider(InputWidget):
    """Useful to create a slider input."""

    type: InputWidgetType = "slider"
    initial: float = 0
    min: float = 0
    max: float = 10
    step: float = 1

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "min": self.min,
            "max": self.max,
            "step": self.step,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Select(InputWidget):
    """Useful to create a select input."""

    type: InputWidgetType = "select"
    initial: Optional[str] = None
    initial_index: Optional[int] = None
    initial_value: Optional[str] = None
    values: List[str] = Field(default_factory=lambda: [])
    items: Dict[str, str] = Field(default_factory=lambda: defaultdict(dict))

    def __post_init__(
        self,
    ) -> None:
        super().__post_init__()

        if not self.values and not self.items:
            raise ValueError("Must provide values or items to create a Select")

        if self.values and self.items:
            raise ValueError(
                "You can only provide either values or items to create a Select"
            )

        if not self.values and self.initial_index is not None:
            raise ValueError(
                "Initial_index can only be used in combination with values to create a Select"
            )

        if self.items:
            self.initial = self.initial_value
        elif self.values:
            self.items = {value: value for value in self.values}
            self.initial = (
                self.values[self.initial_index]
                if self.initial_index is not None
                else self.initial_value
            )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "items": [
                {"label": id, "value": value} for id, value in self.items.items()
            ],
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class TextInput(InputWidget):
    """Useful to create a text input."""

    type: InputWidgetType = "textinput"
    initial: Optional[str] = None
    placeholder: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "placeholder": self.placeholder,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class NumberInput(InputWidget):
    """Useful to create a number input."""

    type: InputWidgetType = "numberinput"
    initial: Optional[float] = None
    placeholder: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "placeholder": self.placeholder,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Tags(InputWidget):
    """Useful to create an input for an array of strings."""

    type: InputWidgetType = "tags"
    initial: List[str] = Field(default_factory=lambda: [])
    values: List[str] = Field(default_factory=lambda: [])

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "tooltip": self.tooltip,
            "description": self.description,
        }



"""The code for the Agenta SDK"""

import os
import sys
import time
import inspect
import argparse
import traceback
import functools
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, Optional, Tuple, List, Union

from fastapi import Body, FastAPI, UploadFile
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware

import chainlo.agenta_cli.agenta as agenta
from .context import save_context
from .router import router as router
from .types import (
    Context,
    DictInput,
    FloatParam,
    InFile,
    IntParam,
    MultipleChoiceParam,
    TextParam,
    MessagesInput,
    FileInputURL,
    FuncResponse,
    BinaryParam,
)

from chainlo.server import app

# app = FastAPI()

# origins = [
#     "*",
# ]

# app.add_middleware(
#     CORSMiddleware,
#     allow_origins=origins,
#     allow_credentials=True,
#     allow_methods=["*"],
#     allow_headers=["*"],
# )

# app.include_router(router, prefix="")


def ingest_file(upfile: UploadFile):
    temp_file = NamedTemporaryFile(delete=False)
    temp_file.write(upfile.file.read())
    temp_file.close()
    return InFile(file_name=upfile.filename, file_path=temp_file.name)


def entrypoint(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to wrap a function for HTTP POST and terminal exposure.

    Args:
        func: Function to wrap.

    Returns:
        Wrapped function for HTTP POST and terminal.
    """

    endpoint_name = "generate"
    func_signature = inspect.signature(func)
    config_params = agenta.config.all()
    ingestible_files = extract_ingestible_files(func_signature)

    @functools.wraps(func)
    async def wrapper(*args, **kwargs) -> Any:
        func_params, api_config_params = split_kwargs(kwargs, config_params)
        ingest_files(func_params, ingestible_files)
        agenta.config.set(**api_config_params)
        return await execute_function(func, *args, **func_params)

    @functools.wraps(func)
    async def wrapper_deployed(*args, **kwargs) -> Any:
        func_params = {
            k: v for k, v in kwargs.items() if k not in ["config", "environment"]
        }
        if "environment" in kwargs and kwargs["environment"] is not None:
            agenta.config.pull(environment_name=kwargs["environment"])
        elif "config" in kwargs and kwargs["config"] is not None:
            agenta.config.pull(config_name=kwargs["config"])
        else:
            agenta.config.pull(config_name="default")
        return await execute_function(func, *args, **func_params)

    update_function_signature(wrapper, func_signature, config_params, ingestible_files)
    route = f"/{endpoint_name}"
    app.post(route, response_model=FuncResponse)(wrapper)

    update_deployed_function_signature(
        wrapper_deployed,
        func_signature,
        ingestible_files,
    )
    route_deployed = f"/{endpoint_name}_deployed"
    app.post(route_deployed, response_model=FuncResponse)(wrapper_deployed)
    override_schema(
        openapi_schema=app.openapi(),
        func_name=func.__name__,
        endpoint=endpoint_name,
        params={**config_params, **func_signature.parameters},
    )

    if is_main_script(func):
        handle_terminal_run(
            func,
            func_signature.parameters,
            config_params,
            ingestible_files,
        )
    return None


def extract_ingestible_files(
    func_signature: inspect.Signature,
) -> Dict[str, inspect.Parameter]:
    """Extract parameters annotated as InFile from function signature."""

    return {
        name: param
        for name, param in func_signature.parameters.items()
        if param.annotation is InFile
    }


def split_kwargs(
    kwargs: Dict[str, Any], config_params: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Split keyword arguments into function parameters and API configuration parameters."""

    func_params = {k: v for k, v in kwargs.items() if k not in config_params}
    api_config_params = {k: v for k, v in kwargs.items() if k in config_params}
    return func_params, api_config_params


def ingest_files(
    func_params: Dict[str, Any], ingestible_files: Dict[str, inspect.Parameter]
) -> None:
    """Ingest files specified in function parameters."""

    for name in ingestible_files:
        if name in func_params and func_params[name] is not None:
            func_params[name] = ingest_file(func_params[name])


async def execute_function(
    func: Callable[..., Any], *args, **func_params
) -> Union[Dict[str, Any], JSONResponse]:
    """Execute the function and handle any exceptions."""

    try:
        """Note: The following block is for backward compatibility.
        It allows functions to work seamlessly whether they are synchronous or asynchronous.
        For synchronous functions, it calls them directly, while for asynchronous functions,
        it awaits their execution.
        """
        is_coroutine_function = inspect.iscoroutinefunction(func)
        start_time = time.perf_counter()
        if is_coroutine_function:
            result = await func(*args, **func_params)
        else:
            result = func(*args, **func_params)
        end_time = time.perf_counter()
        latency = end_time - start_time

        if isinstance(result, Context):
            save_context(result)
        if isinstance(result, Dict):
            return FuncResponse(**result, latency=round(latency, 4)).dict()
        if isinstance(result, str):
            return FuncResponse(message=result, latency=round(latency, 4)).dict()
    except Exception as e:
        return handle_exception(e)


def handle_exception(e: Exception) -> JSONResponse:
    """Handle exceptions and return a JSONResponse."""

    status_code: int = e.status_code if hasattr(e, "status_code") else 500
    traceback_str = traceback.format_exception(e, value=e, tb=e.__traceback__)
    return JSONResponse(
        status_code=status_code,
        content={"error": str(e), "traceback": "".join(traceback_str)},
    )


def update_wrapper_signature(wrapper: Callable[..., Any], updated_params: List):
    """
    Updates the signature of a wrapper function with a new list of parameters.

    Args:
        wrapper (callable): A callable object, such as a function or a method, that requires a signature update.
        updated_params (List[inspect.Parameter]): A list of `inspect.Parameter` objects representing the updated parameters
            for the wrapper function.
    """

    wrapper_signature = inspect.signature(wrapper)
    wrapper_signature = wrapper_signature.replace(parameters=updated_params)
    wrapper.__signature__ = wrapper_signature


def update_function_signature(
    wrapper: Callable[..., Any],
    func_signature: inspect.Signature,
    config_params: Dict[str, Any],
    ingestible_files: Dict[str, inspect.Parameter],
) -> None:
    """Update the function signature to include new parameters."""

    updated_params = []
    add_config_params_to_parser(updated_params, config_params)
    add_func_params_to_parser(updated_params, func_signature, ingestible_files)
    update_wrapper_signature(wrapper, updated_params)


def update_deployed_function_signature(
    wrapper: Callable[..., Any],
    func_signature: inspect.Signature,
    ingestible_files: Dict[str, inspect.Parameter],
) -> None:
    """Update the function signature to include new parameters."""
    updated_params = []
    add_func_params_to_parser(updated_params, func_signature, ingestible_files)
    for param in [
        "config",
        "environment",
    ]:  # we add the config and environment parameters
        updated_params.append(
            inspect.Parameter(
                param,
                inspect.Parameter.KEYWORD_ONLY,
                default=Body(None),
                annotation=str,
            )
        )
    update_wrapper_signature(wrapper, updated_params)


def add_config_params_to_parser(
    updated_params: list, config_params: Dict[str, Any]
) -> None:
    """Add configuration parameters to function signature."""
    for name, param in config_params.items():
        updated_params.append(
            inspect.Parameter(
                name,
                inspect.Parameter.KEYWORD_ONLY,
                default=Body(param),
                annotation=Optional[type(param)],
            )
        )


def add_func_params_to_parser(
    updated_params: list,
    func_signature: inspect.Signature,
    ingestible_files: Dict[str, inspect.Parameter],
) -> None:
    """Add function parameters to function signature."""
    for name, param in func_signature.parameters.items():
        if name in ingestible_files:
            updated_params.append(
                inspect.Parameter(name, param.kind, annotation=UploadFile)
            )
        else:
            updated_params.append(
                inspect.Parameter(
                    name,
                    inspect.Parameter.KEYWORD_ONLY,
                    default=Body(..., embed=True),
                    annotation=param.annotation,
                )
            )


def is_main_script(func: Callable) -> bool:
    """
    Check if the script containing the function is the main script being run.

    Args:
        func (Callable): The function object to check.

    Returns:
        bool: True if the script containing the function is the main script, False otherwise.

    Example:
        if is_main_script(my_function):
            print("This is the main script.")
    """
    return (
        os.path.splitext(os.path.basename(sys.argv[0]))[0]
        == os.path.splitext(os.path.basename(inspect.getfile(func)))[0]
    )


def handle_terminal_run(
    func: Callable,
    func_params: Dict[str, Any],
    config_params: Dict[str, Any],
    ingestible_files: Dict,
) -> None:
    """
    Parses command line arguments and sets configuration when script is run from the terminal.

    Args:
        func_params (dict): A dictionary containing the function parameters and their annotations.
        config_params (dict): A dictionary containing the configuration parameters.

    Example:
        handle_terminal_run(func_params=inspect.signature(my_function).parameters, config_params=config.all())
    """

    # For required parameters, we add them as arguments
    parser = argparse.ArgumentParser()
    for name, param in func_params.items():
        if name in ingestible_files:
            parser.add_argument(name, type=str)
        else:
            parser.add_argument(name, type=param.annotation)

    for name, param in config_params.items():
        if type(param) is MultipleChoiceParam:
            parser.add_argument(
                f"--{name}",
                type=str,
                default=param.default,
                choices=param.choices,
            )
        else:
            parser.add_argument(
                f"--{name}",
                type=type(param),
                default=param,
            )

    args = parser.parse_args()

    # split the arg list into the arg in the app_param and
    # the args from the sig.parameter
    args_config_params = {k: v for k, v in vars(args).items() if k in config_params}
    args_func_params = {k: v for k, v in vars(args).items() if k not in config_params}
    for name in ingestible_files:
        args_func_params[name] = InFile(
            file_name=Path(args_func_params[name]).stem,
            file_path=args_func_params[name],
        )
    agenta.config.set(**args_config_params)


def override_schema(openapi_schema: dict, func_name: str, endpoint: str, params: dict):
    """
    Overrides the default openai schema generated by fastapi with additional information about:
    - The choices available for each MultipleChoiceParam instance
    - The min and max values for each FloatParam instance
    - The min and max values for each IntParam instance
    - The default value for DictInput instance
    - The default value for MessagesParam instance
    - The default value for FileInputURL instance
    - The default value for BinaryParam instance
    - ... [PLEASE ADD AT EACH CHANGE]

    Args:
        openapi_schema (dict): The openapi schema generated by fastapi
        func_name (str): The name of the function to override
        endpoint (str): The name of the endpoint to override
        params (dict(param_name, param_val)): The dictionary of the parameters for the function
    """

    def find_in_schema(schema: dict, param_name: str, xparam: str):
        """Finds a parameter in the schema based on its name and x-parameter value"""
        for _, value in schema.items():
            value_title_lower = str(value.get("title")).lower()
            value_title = (
                "_".join(value_title_lower.split())
                if len(value_title_lower.split()) >= 2
                else value_title_lower
            )

            if (
                isinstance(value, dict)
                and value.get("x-parameter") == xparam
                and value_title == param_name
            ):
                return value

    schema_to_override = openapi_schema["components"]["schemas"][
        f"Body_{func_name}_{endpoint}_post"
    ]["properties"]
    for param_name, param_val in params.items():
        if isinstance(param_val, MultipleChoiceParam):
            subschema = find_in_schema(schema_to_override, param_name, "choice")
            default = str(param_val)
            param_choices = param_val.choices
            choices = (
                [default] + param_choices
                if param_val not in param_choices
                else param_choices
            )
            subschema["enum"] = choices
            subschema["default"] = default if default in param_choices else choices[0]
        if isinstance(param_val, FloatParam):
            subschema = find_in_schema(schema_to_override, param_name, "float")
            subschema["minimum"] = param_val.minval
            subschema["maximum"] = param_val.maxval
            subschema["default"] = param_val
        if isinstance(param_val, IntParam):
            subschema = find_in_schema(schema_to_override, param_name, "int")
            subschema["minimum"] = param_val.minval
            subschema["maximum"] = param_val.maxval
            subschema["default"] = param_val
        if (
            isinstance(param_val, inspect.Parameter)
            and param_val.annotation is DictInput
        ):
            subschema = find_in_schema(schema_to_override, param_name, "dict")
            subschema["default"] = param_val.default["default_keys"]
        if isinstance(param_val, TextParam):
            subschema = find_in_schema(schema_to_override, param_name, "text")
            subschema["default"] = param_val
        if (
            isinstance(param_val, inspect.Parameter)
            and param_val.annotation is MessagesInput
        ):
            subschema = find_in_schema(schema_to_override, param_name, "messages")
            subschema["default"] = param_val.default
        if (
            isinstance(param_val, inspect.Parameter)
            and param_val.annotation is FileInputURL
        ):
            subschema = find_in_schema(schema_to_override, param_name, "file_url")
            subschema["default"] = "https://example.com"
        if isinstance(param_val, BinaryParam):
            subschema = find_in_schema(schema_to_override, param_name, "bool")
            subschema["default"] = param_val.default
